/* A BLAS compatible interface for matrix-matrix generator
 *
 * This interface compatible with the BLAS DGEMM.
 *
 * Written by Dominic LAM <ctlam@po.eecs.berkeley.edu>
 * 
 * $Id: dgemm.c,v 1.1 1995/10/10 05:51:03 ctlam Exp $
 */

#include <stdio.h>

#define max(a,b)     (((a) > (b)) ? (a) : (b))


extern void mdmd_a1bc_md(int, int, int, double*, double*, double*, 
			 int, int, int, double);
extern void mdmd_acbc_md(int, int, int, double*, double*, double*, 
			 int, int, int, double, double);

extern void mdmdt_a1bc_md(int, int, int, double*, double*, double*, 
			  int, int, int, double);
extern void mdmdt_acbc_md(int, int, int, double*, double*, double*, 
			  int, int, int, double, double);

extern void mdtmd_a1bc_md(int, int, int, double*, double*, double*, 
			  int, int, int, double);
extern void mdtmd_acbc_md(int, int, int, double*, double*, double*, 
			  int, int, int, double, double);

extern void mdtmdt_a1bc_md(int, int, int, double*, double*, double*, 
			   int, int, int, double);
extern void mdtmdt_acbc_md(int, int, int, double*, double*, double*, 
			   int, int, int, double, double);

/* error handling routine */
extern void xerbla(char *, int*);

void
dgemm(char* transA, char* transB, int* M, int* N, int* K, double* alpha, double* A, int* Astride, double* B, int* Bstride, double* beta, double* C, int* Cstride){

  int info = 0;

  /* error checking */

  if (*M < 0) {
    info = 3;
    xerbla ("DGEMM ", &info);   /* six char needed in xerbla :-( */
    return;
  }
  else if (*N < 0) {
    info = 4;
    xerbla ("DGEMM ", &info);   
    return;
  }
  else if (*K < 0) {
    info = 5;
    xerbla ("DGEMM ", &info);   
    return;
  }
  
  if (*transA == 'n' || *transA == 'N'){
    if (*transB == 'n' || *transB == 'N'){

      /* error checking */
      if (*Astride < max(1,*M)){
	info = 8;
	xerbla ("DGEMM ", &info);   
      } else if (*Bstride < max(1,*K)) {
	info = 10;
	xerbla ("DGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla ("DGEMM ", &info);   
      }

      /* C = AB + beta*C */
      else if (*alpha == 1) 
	mdmd_a1bc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);
      /* C = alpha*AB + beta*C */
      else 
	mdmd_acbc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);
    } else {

      /* error checking */
      if (*transB != 'c' && *transB != 'C' && 
	  *transB != 't' && *transB != 'T'){
	info = 2;
	xerbla ("DGEMM ", &info);   
      } else if (*Astride < max(1,*M)) {
	info = 8;
	xerbla ("DGEMM ", &info);   
      } else if (*Bstride < max(1,*N)) {
	info = 10;
	xerbla ("DGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla ("DGEMM ", &info);   
      }
      /* C = AB' + beta*C */
      else if (*alpha == 1)
	mdtmd_a1bc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);
      /* C = alpha*AB' + beta*C */
      else 
	mdtmd_acbc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);
    }
  } else {
    if (*transB == 'n' || *transB == 'N'){

      /* error checking */
      if (*transA != 'c' && *transA != 'C' &&
	  *transA != 't' && *transA != 'T') {
	info = 1;
	xerbla ("DGEMM ", &info);   
      } else if (*Astride < max(1,*K)) {
	info = 8;
	xerbla ("DGEMM ", &info);   
      } else if (*Bstride < max(1,*K)) {
	info = 10;
	xerbla ("DGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla ("DGEMM ", &info);   
      }

      /* C = A'B + beta*C */
      else if (*alpha == 1)
	mdmdt_a1bc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);

      /* C = alpha*A'B + beta*C */
      else 
	mdmdt_acbc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);

    } else {

      /* error checking */
      if (*transA != 'c' && *transA != 'C' &&
	  *transA != 't' && *transA != 'T') {
	info = 1;
	xerbla ("DGEMM ", &info);   
      } else if (*transB != 'c' && *transB != 'C' &&
	       *transB != 't' && *transB != 'T') {
	info = 2;
	xerbla ("DGEMM ", &info);   
      } else if (*Astride < max(1,*K)) {
	info = 8;
	xerbla ("DGEMM ", &info);   
      } else if (*Bstride < max(1,*N)) {
	info = 10;
	xerbla ("DGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla ("DGEMM ", &info);   
      }

      /* C = A'B' * beta*C */
      else if (*alpha == 1)
	mdtmdt_a1bc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);
      /* C = alpha*A'B' * beta*C */
      else 
	mdtmdt_acbc_md
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);
    }
  }
}



