/* A BLAS compatible interface for matrix-matrix generator
 *
 * This interface compatible with the BLAS SGEMM.
 *
 * Written by Dominic LAM <ctlam@po.eecs.berkeley.edu>
 * 
 * $Id: sgemm.c,v 1.1 1995/10/10 05:51:07 ctlam Exp $
 */

#include <stdio.h>

#define max(a,b)     (((a) > (b)) ? (a) : (b))


extern void mfmf_a1bc_mf(int, int, int, float*, float*, float*, 
			 int, int, int, float);
extern void mfmf_acbc_mf(int, int, int, float*, float*, float*, 
			 int, int, int, float, float);

extern void mfmft_a1bc_mf(int, int, int, float*, float*, float*, 
			  int, int, int, float);
extern void mfmft_acbc_mf(int, int, int, float*, float*, float*, 
			  int, int, int, float, float);

extern void mftmf_a1bc_mf(int, int, int, float*, float*, float*, 
			  int, int, int, float);
extern void mftmf_acbc_mf(int, int, int, float*, float*, float*, 
			  int, int, int, float, float);

extern void mftmft_a1bc_mf(int, int, int, float*, float*, float*, 
			   int, int, int, float);
extern void mftmft_acbc_mf(int, int, int, float*, float*, float*, 
			   int, int, int, float, float);

/* error handling routine */
extern void xerbla_(char *, int*);

void
sgemm_(char* transA, char* transB, int* M, int* N, int* K, float* alpha, float* A, int* Astride, float* B, int* Bstride, float* beta, float* C, int* Cstride){

  int info = 0;

  /* error checking */

  if (*M < 0) {
    info = 3;
    xerbla_ ("SGEMM ", &info);   /* six char needed in xerbla_ :-( */
    return;
  }
  else if (*N < 0) {
    info = 4;
    xerbla_ ("SGEMM ", &info);   
    return;
  }
  else if (*K < 0) {
    info = 5;
    xerbla_ ("SGEMM ", &info);   
    return;
  }
  
  if (*transA == 'n' || *transA == 'N'){
    if (*transB == 'n' || *transB == 'N'){

      /* error checking */
      if (*Astride < max(1,*M)){
	info = 8;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Bstride < max(1,*K)) {
	info = 10;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla_ ("SGEMM ", &info);   
      }

      /* C = AB + beta*C */
      else if (*alpha == 1) 
	mfmf_a1bc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);
      /* C = alpha*AB + beta*C */
      else 
	mfmf_acbc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);
    } else {

      /* error checking */
      if (*transB != 'c' && *transB != 'C' && 
	  *transB != 't' && *transB != 'T'){
	info = 2;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Astride < max(1,*M)) {
	info = 8;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Bstride < max(1,*N)) {
	info = 10;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla_ ("SGEMM ", &info);   
      }
      /* C = AB' + beta*C */
      else if (*alpha == 1)
	mftmf_a1bc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);
      /* C = alpha*AB' + beta*C */
      else 
	mftmf_acbc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);
    }
  } else {
    if (*transB == 'n' || *transB == 'N'){

      /* error checking */
      if (*transA != 'c' && *transA != 'C' &&
	  *transA != 't' && *transA != 'T') {
	info = 1;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Astride < max(1,*K)) {
	info = 8;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Bstride < max(1,*K)) {
	info = 10;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla_ ("SGEMM ", &info);   
      }

      /* C = A'B + beta*C */
      else if (*alpha == 1)
	mfmft_a1bc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);

      /* C = alpha*A'B + beta*C */
      else 
	mfmft_acbc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);

    } else {

      /* error checking */
      if (*transA != 'c' && *transA != 'C' &&
	  *transA != 't' && *transA != 'T') {
	info = 1;
	xerbla_ ("SGEMM ", &info);   
      } else if (*transB != 'c' && *transB != 'C' &&
	       *transB != 't' && *transB != 'T') {
	info = 2;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Astride < max(1,*K)) {
	info = 8;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Bstride < max(1,*N)) {
	info = 10;
	xerbla_ ("SGEMM ", &info);   
      } else if (*Cstride < max(1,*M)) {
	info = 13;
	xerbla_ ("SGEMM ", &info);   
      }

      /* C = A'B' * beta*C */
      else if (*alpha == 1)
	mftmft_a1bc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *beta);
      /* C = alpha*A'B' * beta*C */
      else 
	mftmft_acbc_mf
	  (*N, *K, *M, B, A, C, *Bstride, *Astride, *Cstride, *alpha, *beta);
    }
  }
}



