/*
    $Id: IPM_timer.c,v 1.24 1995/09/12 00:18:55 krste Exp $

    Interval Performance Measurement package.
    Provides a portable interface to extract performance measurements over
    some interval of a program's execution.
*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/******************************************************************/
/* Configure must set up these symbols depending on system type.  */
/******************************************************************/

/* Must define at least one module. */
#if defined(HAVE_GETRUSAGE)
#elif defined(HAVE_GETTIMEOFDAY)
#elif defined(HAVE_SYSSGI)
#elif defined(HAVE_GETHRTIME)
#elif defined(HAVE_GETTIMER)
#elif defined(HAVE_T0VCOUNT)
#else
error
#endif

/* Must have defined the default module. */

#ifndef IPM_init_DEFAULT
error
#endif

/* Include IPM header files after configured symbols. */

#include "IPM_timer.h"

#ifdef HAVE_GETRUSAGE
#include "IPM_timer_getrusage.h"
#endif

#ifdef HAVE_GETTIMEOFDAY
#include "IPM_timer_gettimeofday.h"
#endif

#ifdef HAVE_T0VCOUNT
#include "IPM_timer_t0vcount.h"
#endif

#ifdef HAVE_SYSSGI
#include "IPM_timer_syssgi.h"
#endif

#ifdef HAVE_GETHRTIME
#include "IPM_timer_gethrtime.h"
#endif

#ifdef HAVE_GETTIMER
#include "IPM_timer_gettimer.h"
#endif

/* Initialization function parses environment, sets up selected timing
   module in dispatch table, and calls modules initialization code. */

static void
IPM_init(void)
{
    static int already_initialized = 0;
    const char *mp;

    /* If we execute this code twice, the timer module must not have
       initialized all the function pointers correctly. */
    assert(already_initialized==0);
    already_initialized = 1;

    mp = getenv("IPM_TIMER_MODULE");

    if (!mp)
        IPM_init_DEFAULT();
    else

#ifdef HAVE_GETRUSAGE
    if (!strcmp(mp, "GETRUSAGE"))
        IPM_init_getrusage();
    else
#endif

#ifdef HAVE_GETTIMEOFDAY
    if (!strcmp(mp, "GETTIMEOFDAY"))
        IPM_init_gettimeofday();
    else
#endif

#ifdef HAVE_T0VCOUNT
    if (!strcmp(mp, "T0VCOUNT"))
        IPM_init_t0vcount();
    else
#endif

#ifdef HAVE_SYSSGI
    if (!strcmp(mp, "SYSSGI"))
        IPM_init_syssgi();
    else
#endif

#ifdef HAVE_GETHRTIME
    if (!strcmp(mp, "GETHRTIME"))
        IPM_init_gethrtime();
    else
#endif

#ifdef HAVE_GETTIMER
    if (!strcmp(mp, "GETTIMER"))
        IPM_init_gettimer();
    else
#endif

    {
        fprintf(stderr, "Error: IPM_TIMER_MODULE \"%s\" unavailable.\n", mp);
        exit(1);
    }
}

static void
IPM_abort(const char *message)
{
    fprintf(stderr, "IPM aborting: %s\n", message);
    fflush(stderr);
    abort();
}

double
IPM_timer_resolution_init(void)
{
    IPM_init();                    /* Set things up, */
    return IPM_timer_resolution(); /* then try again. */
}

double
IPM_timer_range_init(void)
{
    IPM_init();               /* Set things up, */
    return IPM_timer_range(); /* then try again. */
}
    

double
IPM_timer_read_init(const IPM_timer* t)
{
    IPM_abort("IPM_timer_read() called before measurement.");
    return 0.0;             /* Keep compiler happy. */
}

void
IPM_timer_report_init(const char* tag, const IPM_timer* t, char* buf)
{
    IPM_abort("IPM_timer_report() called before measurement.");
}

void
IPM_timer_clear_init(IPM_timer* t)
{
    IPM_init();             /* Set things up, */
    IPM_timer_clear(t);     /* then try again. */
}

void
IPM_timer_start_init(IPM_timer* t)
{
    IPM_init();             /* Set things up, */
    IPM_timer_start(t);     /* then try again. */
}

void
IPM_timer_stop_init(IPM_timer* t)
{
    IPM_abort("Called IPM_stop() before IPM_start().");
}

double
IPM_timer_get_time_init(void)
{
    IPM_init();                  /* Set things up, */
    return IPM_timer_get_time(); /* then try again. */
}

/*
    Initialize the dispatch table to point to the initialize/abort
    functions.
*/

IPM_timer_resolution_p_t IPM_timer_resolution_p = IPM_timer_resolution_init;
IPM_timer_range_p_t IPM_timer_range_p = IPM_timer_range_init;
IPM_timer_clear_p_t IPM_timer_clear_p = IPM_timer_clear_init;
IPM_timer_start_p_t IPM_timer_start_p = IPM_timer_start_init;
IPM_timer_stop_p_t IPM_timer_stop_p = IPM_timer_stop_init;
IPM_timer_read_p_t IPM_timer_read_p = IPM_timer_read_init;
IPM_timer_report_p_t IPM_timer_report_p = IPM_timer_report_init;
IPM_timer_get_time_p_t IPM_timer_get_time_p = IPM_timer_get_time_init;

