const char* IPM_timer_gethrtime_rcsid = "$Header: /u/spert/src/IPM/RCS/IPM_timer_gethrtime.c,v 1.17 1995/09/12 00:52:20 krste Exp $";

/*********************************************************************/
/* Code to use the gethrtime()/gethrvtime() calls on SunOS5 systems. */
/*********************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>

/********************************************************/
/* Configure can set up these system dependent values.  */
/********************************************************/

#ifndef IPM_GETHRTIME_RANGE
   /* 64-bit counter with 1 ns resolution. */
#  define IPM_GETHRTIME_RANGE (18446744073.6)
#endif

#ifndef IPM_GETHRTIME_RESOLUTION
   /* Rather than dying, substitute a sensible value. */
   /* Apparently, this is the resolution on most Sun systems. */
#  define IPM_GETHRTIME_RESOLUTION (500e-9)
#endif

#ifndef IPM_GETHRVTIME_RESOLUTION
   /* 100 Hz quanta on SunOS 5 SPARC systems. */
#  define IPM_GETHRVTIME_RESOLUTION (0.010)
#endif

#include "IPM_timer.h"
#include "IPM_timer_gethrtime.h"

typedef struct
{
    hrtime_t start;
    unsigned long n_intervals;
    hrtime_t total;
} IPM_timer_gethrtime;

static double resolution;       /* Store resolution. */

static double
IPM_timer_resolution_gethrtime(void)
{ return resolution; }

static double
IPM_timer_range_gethrtime(void)
{ return IPM_GETHRTIME_RANGE; }

static void
IPM_timer_clear_gethrtime(IPM_timer *ipmt)
{
    IPM_timer_gethrtime *ghrt = (IPM_timer_gethrtime *) ipmt;
    
    ghrt->n_intervals = 0;
    ghrt->total = 0;
}

static void
IPM_timer_start_gethrtime(IPM_timer *ipmt)
{
    IPM_timer_gethrtime *ghrt = (IPM_timer_gethrtime *) ipmt;

    ghrt->start = gethrtime();
}

static void
IPM_timer_stop_gethrtime(IPM_timer *ipmt)
{
    IPM_timer_gethrtime *ghrt = (IPM_timer_gethrtime *) ipmt;

    const hrtime_t stop = gethrtime();
    
    ghrt->n_intervals++;
    ghrt->total += (stop - ghrt->start);
}

static void
IPM_timer_start_gethrvtime(IPM_timer *ipmt)
{
    IPM_timer_gethrtime *ghrt = (IPM_timer_gethrtime *) ipmt;

    ghrt->start = gethrvtime();
}

static void
IPM_timer_stop_gethrvtime(IPM_timer *ipmt)
{
    IPM_timer_gethrtime *ghrt = (IPM_timer_gethrtime *) ipmt;

    const hrtime_t stop = gethrvtime();
    
    ghrt->n_intervals++;
    ghrt->total += (stop - ghrt->start);
}

static double
IPM_timer_read_gethrtime(const IPM_timer *ipmt)
{
    const IPM_timer_gethrtime *ghrt = (const IPM_timer_gethrtime *) ipmt;

    /* TODO: Subtract calling overhead. */
    return ghrt->total * 1e-9;
}

static void
IPM_timer_report_gethrtime(const char *tag,
                 const IPM_timer *ipmt,
                 char *buf)
{
    /* TODO: This makes assumptions about resolution. */
    (void) sprintf(buf, "INTERVAL: %s\nREAL_TIME: %13.7f\n",
                   tag,
                   IPM_timer_read_gethrtime(ipmt));
}

static void
IPM_timer_report_gethrvtime(const char *tag,
                            const IPM_timer *ipmt,
                            char *buf)
{
    /* TODO: This makes assumptions about resolution. */
    sprintf(buf, "INTERVAL: %s\nPROCESS_TIME: %13.2f\n",
            tag,
            IPM_timer_read_gethrtime(ipmt));
}

static double
IPM_timer_get_time_gethrtime(void)
{
    hrtime_t t;

    t = gethrtime();

    return t * 1e-9;
}

static double
IPM_timer_get_time_gethrvtime(void)
{
    hrtime_t t;

    t = gethrvtime();

    return t * 1e-9;
}

void
IPM_init_gethrtime(void)
{
    /* Parse environment variables. */
    const char *timer_str = getenv("IPM_TIMER_TYPE");

    if (!timer_str || !strcmp(timer_str, "REAL"))
    {
        struct timespec res;

#ifdef HAVE_LIBPOSIX4
        if (clock_getres(CLOCK_REALTIME, &res))
        {
            perror("IPM_init_gethrtime():");
            abort();
        }
        resolution = (res.tv_sec * 1e9 + res.tv_nsec) * 1e-9;
#else
        resolution = IPM_GETHRTIME_RESOLUTION;
#endif

        IPM_timer_start_p = IPM_timer_start_gethrtime;
        IPM_timer_stop_p = IPM_timer_stop_gethrtime;
        IPM_timer_report_p = IPM_timer_report_gethrtime;
        IPM_timer_get_time_p = IPM_timer_get_time_gethrtime;
    }
    else if (!strcmp(timer_str, "PROCESS"))
    {
        resolution = IPM_GETHRVTIME_RESOLUTION;

        IPM_timer_start_p = IPM_timer_start_gethrvtime;
        IPM_timer_stop_p = IPM_timer_stop_gethrvtime;
        IPM_timer_report_p = IPM_timer_report_gethrvtime;
        IPM_timer_get_time_p = IPM_timer_get_time_gethrvtime;
    }
    else
    {
        fprintf(stderr,
                "IPM_init() : No such timer type: %s\n", timer_str);
        exit(1);
    }

    /* TODO: Get calling overhead from arguments. */

    /* Make sure opaque struct large enough. */
    assert(sizeof(IPM_timer_gethrtime) <= IPM_TIMER_STRUCT_MAX_SIZE);

    /* Install remaining function pointers. */
    IPM_timer_resolution_p = IPM_timer_resolution_gethrtime;
    IPM_timer_range_p = IPM_timer_range_gethrtime;
    IPM_timer_clear_p = IPM_timer_clear_gethrtime;
    IPM_timer_read_p = IPM_timer_read_gethrtime;
}
