const char* IPM_timer_getrusage_rcsid = "$Header: /u/spert/src/IPM/RCS/IPM_timer_getrusage.c,v 1.15 1995/09/12 00:13:22 krste Exp $";

/**********************************************************/
/* Code that uses BSD getrusage() to implement the timer. */
/**********************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <sys/time.h>
#include <sys/resource.h>

/******************************************************************/
/* Configure can set up these symbols depending on system type.  */
/******************************************************************/

/* Sun3 has 50Hz clock, other getrusages have 100Hz clock. */
#ifndef IPM_GETRUSAGE_RESOLUTION
#  define IPM_GETRUSAGE_RESOLUTION (0.010)
#endif

/* 2^31-1 is maximum number of +ve seconds in 32-bit timeval. */
#ifndef IPM_GETRUSAGE_RANGE
#  define IPM_GETRUSAGE_RANGE (2147483647.0)
#endif

#include "IPM_timer.h"
#include "IPM_timer_getrusage.h"

typedef struct
{
    struct rusage start;        /* Start values. */
    struct rusage stop;         /* Stop values. */
    double n_intervals;         /* Number of intervals measured. */
    double t_us;                /* Accumulated microseconds, stored in */
                                /* double. */
} IPM_timer_getrusage;

static double
IPM_timer_resolution_getrusage(void)
{
    return IPM_GETRUSAGE_RESOLUTION;
}

static double
IPM_timer_range_getrusage(void)
{
    return IPM_GETRUSAGE_RANGE;
}

static void
IPM_abort_getrusage(const char* message)
{
    perror(message);
    abort();
}

static void
IPM_timer_clear_getrusage(IPM_timer* ipmt)
{
    IPM_timer_getrusage *grut = (IPM_timer_getrusage *) ipmt;

    grut->n_intervals = 0.0;       /* Clear number of intervals. */
    grut->t_us = 0.0;              /* Clear accumulated time. */
}

static void
IPM_timer_start_getrusage(IPM_timer* ipmt)
{
    IPM_timer_getrusage *grut = (IPM_timer_getrusage *) ipmt;

    if (getrusage(RUSAGE_SELF, &(grut->start)))
        IPM_abort_getrusage("IPM_timer_start_getrusage()");
}

/* Do timeval calculations in microseconds as this avoids the accuracy loss */
/* that would occur when converting 1e-6 to binary if working in seconds. */
#define TV2US(timeval) (((timeval).tv_sec * 1e6) + (timeval).tv_usec)

static void
IPM_timer_stop_getrusage_process(IPM_timer *ipmt)
{
    IPM_timer_getrusage *grut = (IPM_timer_getrusage *) ipmt;

    if (getrusage(RUSAGE_SELF, &(grut->stop)))
        IPM_abort_getrusage("IPM_timer_stop_getrusage()");

    grut->n_intervals += 1.0;

    /* Add user and system time together. */
    grut->t_us += TV2US(grut->stop.ru_utime) - TV2US(grut->start.ru_utime);
    grut->t_us += TV2US(grut->stop.ru_stime) - TV2US(grut->start.ru_stime);
}

static void
IPM_timer_stop_getrusage_user(IPM_timer *ipmt)
{
    IPM_timer_getrusage *grut = (IPM_timer_getrusage *) ipmt;

    if (getrusage(RUSAGE_SELF, &(grut->stop)))
        IPM_abort_getrusage("IPM_timer_stop_getrusage()");

    grut->n_intervals += 1.0;

    /* Just add on user time. */
    grut->t_us += TV2US(grut->stop.ru_utime) - TV2US(grut->start.ru_utime);
}

static void
IPM_timer_stop_getrusage_system(IPM_timer *ipmt)
{
    IPM_timer_getrusage *grut = (IPM_timer_getrusage *) ipmt;

    if (getrusage(RUSAGE_SELF, &(grut->stop)))
        IPM_abort_getrusage("IPM_timer_stop_getrusage()");

    grut->n_intervals += 1.0;

    /* Just add on system time. */
    grut->t_us += TV2US(grut->stop.ru_stime) - TV2US(grut->start.ru_stime);
}

static double
IPM_timer_read_getrusage(const IPM_timer *ipmt)
{
    const IPM_timer_getrusage *grut = (const IPM_timer_getrusage *) ipmt;

    /* TODO: Subtract off calling overhead. */
    return grut->t_us * 1e-6;
}

static void
IPM_timer_report_getrusage_process(const char *tag,
                                   const IPM_timer *ipmt,
                                   char *buf)
{
    sprintf(buf, "INTERVAL: %s\nPROCESS_TIME: %13.2f\n",
            tag,
            IPM_timer_read_getrusage(ipmt));
}

static void
IPM_timer_report_getrusage_user(const char *tag,
                                const IPM_timer *ipmt,
                                char *buf)
{
    sprintf(buf, "INTERVAL: %s\nUSER_TIME: %13.2f\n",
            tag,
            IPM_timer_read_getrusage(ipmt));
}

static void
IPM_timer_report_getrusage_system(const char *tag,
                                  const IPM_timer *ipmt,
                                  char *buf)
{
    sprintf(buf, "INTERVAL: %s\nSYSTEM_TIME: %13.2f\n",
            tag,
            IPM_timer_read_getrusage(ipmt));
}

static double
IPM_timer_get_time_getrusage_process(void)
{
    struct rusage r;

    if (getrusage(RUSAGE_SELF, &r))
        IPM_abort_getrusage("IPM_timer_get_time_getrusage()");

    return (TV2US(r.ru_utime) + TV2US(r.ru_stime)) * 1e-6;
}

static double
IPM_timer_get_time_getrusage_user(void)
{
    struct rusage r;

    if (getrusage(RUSAGE_SELF, &r))
        IPM_abort_getrusage("IPM_timer_get_time_getrusage()");

    return TV2US(r.ru_utime) * 1e-6;
}

static double
IPM_timer_get_time_getrusage_system(void)
{
    struct rusage r;

    if (getrusage(RUSAGE_SELF, &r))
        IPM_abort_getrusage("IPM_timer_get_time_getrusage()");

    return TV2US(r.ru_stime) * 1e-6;
}

void
IPM_init_getrusage(void)
{
    /* Parse timer type from environment. */
    const char *tp = getenv("IPM_TIMER_TYPE");

    if (!tp || !strcmp(tp, "PROCESS"))
    {
      IPM_timer_stop_p = IPM_timer_stop_getrusage_process;
      IPM_timer_report_p = IPM_timer_report_getrusage_process;
      IPM_timer_get_time_p = IPM_timer_get_time_getrusage_process;
      
    }
    else if (!strcmp(tp, "USER"))
    {
        IPM_timer_stop_p = IPM_timer_stop_getrusage_user;
        IPM_timer_report_p = IPM_timer_report_getrusage_user;
        IPM_timer_get_time_p = IPM_timer_get_time_getrusage_user;
    }
    else if (!strcmp(tp, "SYSTEM"))
    {
        IPM_timer_stop_p = IPM_timer_stop_getrusage_system;
        IPM_timer_report_p = IPM_timer_report_getrusage_system;
        IPM_timer_get_time_p = IPM_timer_get_time_getrusage_system;
    }
    else
    {
        fprintf(stderr, "IPM_init_getrusage(): IPM_TIMER_TYPE \"%s\" unavailable.\n", tp);
        exit(1);
    }

    /* Install remaining function pointers. */
    IPM_timer_resolution_p = IPM_timer_resolution_getrusage;
    IPM_timer_range_p = IPM_timer_range_getrusage;
    IPM_timer_clear_p = IPM_timer_clear_getrusage;
    IPM_timer_start_p = IPM_timer_start_getrusage;
    IPM_timer_read_p = IPM_timer_read_getrusage;
  
    /* Make sure opaque struct large enough. */
    assert(sizeof(IPM_timer_getrusage) <= IPM_TIMER_STRUCT_MAX_SIZE);

    /* TODO: Get calling overhead from arguments. */
}
