const char* IPM_timer_gettimeofday_rcsid = "$Header: /u/spert/src/IPM/RCS/IPM_timer_gettimeofday.c,v 1.12 1995/09/12 00:31:38 krste Exp $";

/*********************************************************/
/* Code that uses gettimeofday() to implement the timer. */
/*********************************************************/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/time.h>
#include <sys/resource.h>

/*****************************************************************/
/* Configure can set up these symbols depending on system type.  */
/*****************************************************************/

/* Some systems have very accurate time of day clocks. */
#ifndef IPM_GETTIMEOFDAY_RESOLUTION
#  define IPM_GETTIMEOFDAY_RESOLUTION (1e-6)
#endif

/* 2^31-1 is maximum number of +ve seconds in a 32-bit timeval. */
#ifndef IPM_GETTIMEOFDAY_RANGE
# define IPM_GETTIMEOFDAY_RANGE (2147483647.0)
#endif

#include "IPM_timer.h"
#include "IPM_timer_gettimeofday.h"

static double
IPM_timer_resolution_gettimeofday(void)
{
    return IPM_GETTIMEOFDAY_RESOLUTION;
}

static double
IPM_timer_range_gettimeofday(void)
{
    return IPM_GETTIMEOFDAY_RANGE;
}

typedef struct
{
    struct timeval start;       /* Start values. */
    struct timeval stop;        /* Stop values. */
    double n_intervals;         /* Number of intervals measured. */
    double t_us;                /* Accumulated microseconds, stored in */
                                /* double. */
} IPM_timer_gettimeofday;

static void
IPM_abort_gettimeofday(const char* message)
{
    perror(message);
    abort();
}

static void
IPM_timer_clear_gettimeofday(IPM_timer *ipmt)
{
    IPM_timer_gettimeofday *gtodt = (IPM_timer_gettimeofday *) ipmt;

    gtodt->n_intervals = 0.0;       /* Clear number of intervals. */
    gtodt->t_us = 0.0;              /* Clear accumulated time. */
}

static void
IPM_timer_start_gettimeofday(IPM_timer *ipmt)
{
    IPM_timer_gettimeofday *gtodt = (IPM_timer_gettimeofday *) ipmt;

    if (gettimeofday(&(gtodt->start), NULL))
        IPM_abort_gettimeofday("IPM_timer_start_gettimeofday()");
}

/* Do timeval calculations in microseconds as this avoids the accuracy loss */
/* that would occur when converting 1e-6 to binary if working in seconds. */
#define TV2US(timeval) (((timeval).tv_sec * 1e6) + (timeval).tv_usec)

static void
IPM_timer_stop_gettimeofday(IPM_timer *ipmt)
{
    IPM_timer_gettimeofday *gtodt = (IPM_timer_gettimeofday *) ipmt;
  
    if (gettimeofday(&(gtodt->stop), NULL))
        IPM_abort_gettimeofday("IPM_timer_stop_gettimeofday()");

    gtodt->n_intervals += 1.0;
    gtodt->t_us += TV2US(gtodt->stop) - TV2US(gtodt->start);
}

static double
IPM_timer_read_gettimeofday(const IPM_timer *ipmt)
{
    const IPM_timer_gettimeofday *gtodt
        = (const IPM_timer_gettimeofday *) ipmt;

    /* TODO: Subtract off calling overhead. */
    return gtodt->t_us * 1e-6;
}

static void
IPM_timer_report_gettimeofday(const char *tag,
                 const IPM_timer *ipmt,
                 char *buf)
{
    const char *timer_name;

    timer_name = "REAL";

    sprintf(buf, "INTERVAL: %s\n%s_TIME: %13.6f\n",
            tag,
            timer_name,
            IPM_timer_read_gettimeofday(ipmt));
}

static double
IPM_timer_get_time_gettimeofday(void)
{
    struct timeval t;

    if (gettimeofday(&t, NULL))
        IPM_abort_gettimeofday("IPM_timer_get_time_gettimeofday()");

    return TV2US(t) * 1e-6;
}

void
IPM_init_gettimeofday(void)
{
    /* Make sure opaque struct large enough. */
    assert(sizeof(IPM_timer_gettimeofday) <= IPM_TIMER_STRUCT_MAX_SIZE);

    /* Install function pointers. */
    IPM_timer_resolution_p = IPM_timer_resolution_gettimeofday;
    IPM_timer_range_p = IPM_timer_range_gettimeofday;
    IPM_timer_clear_p = IPM_timer_clear_gettimeofday;
    IPM_timer_start_p = IPM_timer_start_gettimeofday;
    IPM_timer_stop_p = IPM_timer_stop_gettimeofday;
    IPM_timer_read_p = IPM_timer_read_gettimeofday;
    IPM_timer_report_p = IPM_timer_report_gettimeofday;
    IPM_timer_get_time_p = IPM_timer_get_time_gettimeofday;

    /* TODO: Get calling overhead from arguments. */
}
