/*
** mm_gen: the PHiPAC matrix-matrix multiply generator.
** $Revision: 1.32 $
**
** See the man page mm_gen.1 for details.
**
** Written by: 
**      Jeff Bilmes <bilmes@icsi.berkeley.edu>
**      http://www.icsi.berkeley.edu/~bilmes/phipac
**
** 
** "Copyright (c) 1995 The Regents of the University of California.  All
** rights reserved."  Permission to use, copy, modify, and distribute
** this software and its documentation for any purpose, without fee, and
** without written agreement is hereby granted, provided that the above
** copyright notice and the following two paragraphs appear in all copies
** of this software.
** 
** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
** DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
** OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
** CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**  
** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
** AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
** ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
** PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
** 
*/


static char rcsid[] = "$Header: /da/bilmes/phipac/mm_gen/RCS/mm_gen.c,v 1.32 1995/12/12 08:38:24 bilmes Exp $";
static char rcsrevision[] = "$Revision: 1.32 $";


#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <ctype.h>
#include <math.h>

#include <assert.h>

#define MIN(a,b) ((a)<(b)?(a):(b))

/* command line arguments */
char *outfile=NULL;
FILE *outf = NULL;

char routine_name[BUFSIZ] = { '\0' };

#define MAX_MEM_HIER_DEPTH 20 /* this seems reasonable for
				 the near term at least */
typedef struct cache_block {
  int M,K,N;
  int cumM,cumK,cumN;    /* ascending cumulative blocking */
} CacheBlock;

/*
** Note that the number of levels of cache blocking is
** equal to cache_blocking_entries.
*/
int cache_blocking_entries = 0;
CacheBlock cache_blocking[MAX_MEM_HIER_DEPTH];

typedef enum { True = 1, False = 0 } bool;

typedef enum {     Zero, /* the value is fixed at zero */
		   One,  /* the value is fixed at one */
		   Fix,  /* the value is fixed at some constant */
		   Arb   /* the value is parameterized  */
	       } scaleType;


/* the kind of blocking routine. Either: */
typedef enum {   Fixed,  /* fixed blocked routine */
		 Arb_k,  /* m,n fixed, k arbitrary */
		 Arb_all /* m,n, and k arbitrary */
	       } routineKind;

typedef enum { Normal, Transpose } op;
op opA = Normal;
op opB = Normal;


/* True if we do the register block fringes, False if we don't. */
bool do_fringes=True;

scaleType alpha = One, beta = One;
char *alpha_val = "1.0", *beta_val = "1.0";


bool  sepaccum = False; /* true if we use a separate
			   static memory area to compute
			   the destination matrix first
			   and then copy into the real
			   destination. This is useful
			   for 1) alpha != 1 and 2)
			   when the destination is
			   a separate (higher prec)
			   type. */


bool opt_alpha = False; /* If True, the emitted code
			  will check for when alpha
			  (when present) is 0, 1, etc.
			  and optimize accordingly.
			  */
bool opt_beta = False;  /* Same for beta */

/* This becomes true if alpha is not one
   and sepaccum is false */
bool UseZeroScaleAccumStore = False;

/* The following are the C floating point types that will be used.
   They may be safely changed if needed for a particular compiler.  */
const char* fp_types[] = {   
  "float",   /* presumably IEEE single prec */
  "double",  /* presumably IEEE double prec */
  "long double" /* presumably IEEE double extended prec */
};
const char fp_chars[] = { 'f', 'd', 'l' };
#define FP_TYPE_INVALID (-1)
#define FP_TYPE_FLOAT 0
#define FP_TYPE_DOUBLE 1
#define FP_TYPE_LONG_DOUBLE 2
#define FP_TYPE_DEFAULT FP_TYPE_FLOAT
int fp_type_num = FP_TYPE_INVALID;
int fp_atype_num = FP_TYPE_INVALID; /* accumulator type */
int fp_dtype_num = FP_TYPE_INVALID; /* destination type */
const char* fp_type;
const char* fp_atype; /* accumulator type */
const char* fp_dtype; /* destination type */
char  fp_char;
char  fp_achar;
char  fp_dchar;

bool  call_down = True;

#define DEFAULT_SPACE_PER_NEST 3
int nesting_level=0;
int space_per_nest = DEFAULT_SPACE_PER_NEST;
char space_char = ' ';

void nest() { nesting_level++; }
void unnest() { nesting_level--; }
void nspace()
{
  int i;
  for (i=0;i<space_per_nest*nesting_level;i++)
    fprintf(outf,"%c",space_char);
}


/* returns true of arg is a 2^k for some k */
int pow_of_2(int arg)
{
  int i;
  int bits = 0;
  if (arg < 0) return 0;
  for (i=0;i<sizeof(arg)*8;i++) {
    bits += (arg & 0x1);
    if (bits > 1) return 0;
    arg >>= 1;
  }
  return 1;
}

/* integer log2(u) */
int ilog2 (int u)
{
  int   i;
  if (u == 0)
    return (-1);
  for (i = 0; (u & 1) == 0; i++, u >>= 1);
  return (u == 1 ? i : -1);
}

/* returns the highest bit of u */
int highbit( int u )
{
  int res = 1;
  if (u == 0)
    return 0;
  while (u >>= 1) {
    res <<= 1;
  }
  return res;
}



void
usage(char *name)
{
  fprintf(stderr,"usage: %s [OPTIONS]\n",name);
  fprintf(stderr,"where [OPTIONS] include:\n");  

  fprintf(stderr,"Semantics options:\n");
  fprintf(stderr,"  -opA [N|T]                 : A matrix op. Normal|Transpose\n");
  fprintf(stderr,"  -opB [N|T]                 : B matrix op. Normal|Transpose\n");
  fprintf(stderr,"  -nofringes                 : don't generate reg block fringe code\n");
  fprintf(stderr,"  -alpha [<val>|c]           : fix alpha at value <val> or arbitrary\n");
  fprintf(stderr,"  -beta [<val>|c]            : fix beta at value <val> or arbitrary\n");


  fprintf(stderr,"Optimization options:\n");
  fprintf(stderr,"  -cb M K N [ -cb M K N] ... : reg, L1, L2, etc. blocking\n");
  fprintf(stderr,"  -nocalldown                : no check/call down lower routines first\n");
  fprintf(stderr,"  -sepaccum                  : use separate accumulation mem\n");
  fprintf(stderr,"  -optalpha                  : alpha 0/1/arb runtime checks\n");
  fprintf(stderr,"  -optbeta                   : beta 0/1/arb runtime checks\n");

  fprintf(stderr,"Precision options:\n");
  fprintf(stderr,"  -prec [single|double|ldouble]  : Precision\n");
  fprintf(stderr,"  -sprec [single|double|ldouble] : Source Precision\n");
  fprintf(stderr,"  -aprec [single|double|ldouble] : Accumulator Precision\n");
  fprintf(stderr,"  -dprec [single|double|ldouble] : Destination Precision\n");

  fprintf(stderr,"Code generation & Misc. options:\n");
  fprintf(stderr,"  -help                      : Print this message\n"); 
  fprintf(stderr,"  -file name                 : Write to file 'name'\n"); 
  fprintf(stderr,"  -routine_name name         : Name of routines\n");
  fprintf(stderr,"  -spacechar c               : char to use as space\n");
  fprintf(stderr,"  -numspaces i               : spaces per nest\n");
  fprintf(stderr,"  -version                   : print version and exit.\n");
}



void
check_remaining_params(int argc,int i,int n)
{
  if ((i+n+1)>argc) {
    fprintf(stderr,"Error: too few command line parameters.\n");
    exit(-1);
  }
}

void
parse_args(int argc,char*argv[])
{
  int i;
  if (argc == 1) {
    usage(argv[0]);
    exit(-1);
  }
  /* args initialize */
  outf = stdout;
  for (i=1;i<argc;i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(&(argv[i][1]),"file")) {
	check_remaining_params(argc,i,1);
        outfile = argv[++i];
	if (!strcmp(outfile,"-")) {
	  /* do nothing */
	} else {
	  if ((outf = fopen(outfile,"w")) == NULL) {
	    fprintf(stderr,"Unable to open file %s\n",outfile);
	    exit(1);
	  }
	}
      } else if (!strncmp(&(argv[i][1]),"cb",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,3);
	if (cache_blocking_entries < MAX_MEM_HIER_DEPTH) {
	  if (
	    ((cache_blocking[cache_blocking_entries].M = atoi(argv[++i])) <= 0)
	      ||
	    ((cache_blocking[cache_blocking_entries].K = atoi(argv[++i])) <= 0)
	      ||
	    ((cache_blocking[cache_blocking_entries].N = atoi(argv[++i])) <= 0)
	      ) {
	    fprintf(stderr,"Error: Invalid cb blocking number.\n");
	    exit(-1);
	  }
	  cache_blocking_entries++;
	} else {
	  fprintf(stderr,"Error: Maximum memory hierarchy depth of %d exceeded.\n",MAX_MEM_HIER_DEPTH);
	  exit(-1);
	}
      } else if (!strncmp(&(argv[i][1]),"opA",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"N") || !strcmp(argv[i],"n")) { 
	  opA = Normal;
	} else if (!strcmp(argv[i],"T") || !strcmp(argv[i],"t")) { 
	  opA = Transpose;
	} else
	  fprintf(stderr,"Invalid option (%s), ignored\n",argv[i]);
      } else if (!strncmp(&(argv[i][1]),"opB",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"N") || !strcmp(argv[i],"n")) { 
	  opB = Normal;
	} else if (!strcmp(argv[i],"T") || !strcmp(argv[i],"t")) { 
	  opB = Transpose;
	} else
	  fprintf(stderr,"Invalid option (%s), ignored\n",argv[i]);
      } else if (!strncmp(&(argv[i][1]),"alpha",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"c")) {
	  alpha = Arb;
	} else {
	  char *ptr;
	  double tmp;
	  tmp = strtod(argv[i],&ptr);
	  if (ptr == argv[i]) {
	    fprintf(stderr,"Invalid numeric parameter (%s)\n",argv[i]);
	    exit(-1);
	  }
	  alpha_val = argv[i];
	  if (tmp == 1.0)
	    alpha = One;
	  else if (tmp == 0.0) {
	    fprintf(stderr,"Error: Fixed zero value for alpha nulify's operation.\n");
	    exit(-1);	    
	  } else
	    alpha = Fix;
	}
      } else if (!strncmp(&(argv[i][1]),"beta",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	if (!strcmp(argv[i],"c")) {
	  beta = Arb;
	} else {
	  char *ptr;
	  double tmp;
	  tmp = strtod(argv[i],&ptr);
	  if (ptr == argv[i]) {
	    fprintf(stderr,"Invalid numeric parameter (%s)\n",argv[i]);
	    exit(-1);
	  }
	  beta_val = argv[i];
	  if (tmp == 1.0)
	    beta = One;
	  else if (tmp == 0.0)
	    beta = Zero;
	  else 
	    beta = Fix;
	}
      } else if (!strncmp(&(argv[i][1]),"nofringes",strlen(&(argv[i][1])))) {
	do_fringes = False;
      } else if (!strncmp(&(argv[i][1]),"routine_name",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	i++;
	strncpy(routine_name,argv[i],MIN(BUFSIZ,strlen(argv[i])));
      } else if ((!strncmp(&(argv[i][1]),"precision",strlen(&(argv[i][1])))) ||
		 (!strncmp(&(argv[i][1]),"sprecision",strlen(&(argv[i][1]))))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_type_num = FP_TYPE_LONG_DOUBLE;
	else 
	  fprintf(stderr,"Invalid precision (%s), ignored\n",argv[i+1]);
	i++;
      } else if (!strncmp(&(argv[i][1]),"aprecision",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_atype_num = FP_TYPE_LONG_DOUBLE;
	else 
	  fprintf(stderr,"Invalid precision (%s), ignored\n",argv[i+1]);
	i++;
      } else if (!strncmp(&(argv[i][1]),"dprecision",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	if (!strncmp(argv[i+1],"single",strlen(argv[i+1])) ||
	    !strncmp(argv[i+1],"float",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_FLOAT;
	else if (!strncmp(argv[i+1],"double",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_DOUBLE;
	else if (!strncmp(argv[i+1],"ldouble",strlen(argv[i+1])))
	  fp_dtype_num = FP_TYPE_LONG_DOUBLE;
	else 
	  fprintf(stderr,"Invalid precision (%s), ignored\n",argv[i+1]);
	i++;
      } else if (!strncmp(&(argv[i][1]),"spacechar",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	space_char = *argv[++i];
      } else if (!strncmp(&(argv[i][1]),"numspaces",strlen(&(argv[i][1])))) {
	check_remaining_params(argc,i,1);
	space_per_nest = atoi(argv[++i]);
      } else if (!strncmp(&(argv[i][1]),"nocalldown",strlen(&(argv[i][1])))) {
	call_down = False;
      } else if (!strncmp(&(argv[i][1]),"calldown",strlen(&(argv[i][1])))) {
	call_down = True;
      } else if (!strncmp(&(argv[i][1]),"sepaccum",strlen(&(argv[i][1])))) {
	sepaccum = True;
      } else if (!strncmp(&(argv[i][1]),"optalpha",strlen(&(argv[i][1])))) {
	opt_alpha = True;
      } else if (!strncmp(&(argv[i][1]),"optbeta",strlen(&(argv[i][1])))) {
	opt_beta = True;
      } else if (!strncmp(&(argv[i][1]),"help",strlen(&(argv[i][1])))) {
	usage(argv[0]);
	exit(0);
      } else if (!strncmp(&(argv[i][1]),"-help",strlen(&(argv[i][1])))) {
	usage(argv[0]);
	exit(0);
      } else if (!strncmp(&(argv[i][1]),"version",strlen(&(argv[i][1])))) {
	fprintf(stderr,"%s: %s\n",argv[0],rcsrevision);
	exit(0);
      } else  {
        fprintf(stderr,"Error: Unknown or invalid argument (%s)\n",argv[i]);
	exit(-1);
      }
    }
  }

  /* finish initialize & check */
  if (fp_type_num == FP_TYPE_INVALID)
    fp_type_num = FP_TYPE_DEFAULT;
  if (fp_atype_num == FP_TYPE_INVALID)
    fp_atype_num = fp_type_num;
  if (fp_dtype_num == FP_TYPE_INVALID)
    fp_dtype_num = fp_type_num;


  fp_type = fp_types[fp_type_num];
  fp_atype = fp_types[fp_atype_num];
  fp_dtype = fp_types[fp_dtype_num];
  fp_char = fp_chars[fp_type_num];
  fp_achar = fp_chars[fp_atype_num];
  fp_dchar = fp_chars[fp_dtype_num];

  if (cache_blocking_entries == 0) {
    usage(argv[0]);
    exit(-1);
  }
  if (strlen(routine_name) == 0) {
    /* these should never hold more than
       at most a single char and a NULL */
    char axposechar[2];
    char bxposechar[2];    

    if (opA == Normal)
      strcpy(axposechar,"");
    else 
      strcpy(axposechar,"t");

    if (opB == Normal)
      strcpy(bxposechar,"");
    else 
      strcpy(bxposechar,"t");

    sprintf(routine_name,"mul_m%c%sm%c%s_m%c",
	    fp_char,axposechar,
	    fp_char,bxposechar,
	    fp_dchar);
  }
 
  if (cache_blocking_entries == 1) {
    /* set sepaccum to false here. Why? Because it wouldn't help
       performance to create a separate area in memory to store the
       accumulated block because the accumulation registers already
       constitute such a block.
       */
    sepaccum = False;
  }
 
  if (alpha != One && sepaccum == False)
    UseZeroScaleAccumStore = True;
  else
    UseZeroScaleAccumStore = False;

}


void
initialize()
{
  int i;

  cache_blocking[0].cumM = cache_blocking[0].M;
  cache_blocking[0].cumK = cache_blocking[0].K;
  cache_blocking[0].cumN = cache_blocking[0].N;
  for (i=1;i<cache_blocking_entries;i++) {
    cache_blocking[i].cumM = cache_blocking[i].M * cache_blocking[i-1].cumM;
    cache_blocking[i].cumK = cache_blocking[i].K * cache_blocking[i-1].cumK;
    cache_blocking[i].cumN = cache_blocking[i].N * cache_blocking[i-1].cumN;
  }
  i--;
}

void
generate_macro_name(const int M,const int K, const int N,int fp_dtype_num)
{
 char fp_dchar = fp_chars[fp_dtype_num];
 fprintf(outf,"mul_%sm%c%dx%d%sm%c%dx%d_m%c%dx%d",
	  (opA == Normal ? "" : "t"),
	  fp_char,
	  M,K,
	  (opB == Normal ? "" : "t"),
	  fp_char,
	  K,N,
	  fp_dchar,
	  M,N);
}

/*
** Generate a register blocking macro
*/
void
gen_mm_macro(const int M, const int K, const int N,int fp_dtype_num)
{
  int m,k,n;

  nspace();
  fprintf(outf,"/* Fixed M,K,N = %d,%d,%d fully-unrolled matrix matrix multiply. */\n",
	  M,K,N);
  
  nspace();
  fprintf(outf,"#define ");
  generate_macro_name(M,K,N,fp_dtype_num);
  fprintf(outf,"(");
  for (m=0;m<M;m++)
    for (n=0;n<N;n++)
      fprintf(outf,"c%d%d,",m,n);
  if (opA == Normal) {
    for (m=0;m<M;m++)
      fprintf(outf,"A%d,",m);
  } else {
    fprintf(outf,"A,Astride,");
  }
  if (opB == Normal)
    fprintf(outf,"B,Bstride) \\\n");
  else {
    for (n=0;n<N;n++)
      fprintf(outf,"B%d%s",n,(n==(N-1)?") \\\n":","));
  }

  nspace();
  fprintf(outf,"{ \\\n");

  nest();

  if (opA == Transpose && opB == Transpose) {
    nspace();
    fprintf(outf,"register %s ",fp_type);
    for (m=0;m<M;m++)
      fprintf(outf,"_a%d%s",m,(m < (M-1)?",":"; \\\n"));

    nspace();
    fprintf(outf,"register %s ",fp_type);  
    for (n=0;n<N;n++)    
      fprintf(outf,"_b%d%s",n,(n<(N-1)?",":"; \\\n"));

  } else {
    nspace();
    fprintf(outf,"register %s ",fp_type);
    for (n=0;n<N;n++)
      fprintf(outf,"_b%d%s",n,(n < (N-1)?",":"; \\\n"));

    nspace();
    fprintf(outf,"register %s ",fp_type);  
    for (m=0;m<M;m++)    
      fprintf(outf,"_a%d%s",m,(m<(M-1)?",":"; \\\n"));
  }

  nspace(); fprintf(outf,"\\\n");

  if (opA == opB) {
    char l1,l2;
    int L1,L2;
    if (opA == Normal) {
      l1 = 'a'; l2 = 'b';
      L1 = M; L2 = N;
    } else {
      l1 = 'b'; l2 = 'a';
      L1 = N; L2 = M;
    }

    for (k=0;k<K;k++) {

      nspace(); fprintf(outf,"\\\n");    

      nspace();
      for (n=0;n<L2;n++)
	fprintf(outf,"_%c%d = %c[%d]; ",l2,n,toupper(l2),n);
      fprintf(outf,"\\\n");
      nspace();
      fprintf(outf,"%c += %cstride; \\\n",toupper(l2),toupper(l2));

      for(m=0;m<L1;m++) {

	nspace();
	fprintf(outf,"_%c%d = %c%d[%d]; \\\n",l1,m,toupper(l1),m,k);

	nspace();
	if (opA == Normal) {
	  for (n=0;n<L2;n++)
	    fprintf(outf,"c%d%d += _%c%d*_%c%d; ",m,n,l1,m,l2,n);
	} else {
	  for (n=0;n<L2;n++)
	    fprintf(outf,"c%d%d += _%c%d*_%c%d; ",n,m,l1,m,l2,n);
	}

	fprintf(outf,"\\\n");
      }
    }
  } else {
    for (k=0;k<K;k++) {

      nspace(); fprintf(outf,"\\\n");    

      nspace();
      if (opB == Normal) {
	for (n=0;n<N;n++)
	  fprintf(outf,"_b%d = B[%d]; ",n,n);
	fprintf(outf,"\\\n");
	nspace();
	fprintf(outf,"B += Bstride; \\\n");
      } else {
	for (n=0;n<N;n++)
	  fprintf(outf,"_b%d = B%d[%d]; ",n,n,k);
	fprintf(outf,"\\\n");
      }

      for(m=0;m<M;m++) {
	nspace();
	if (opA == Normal) 
	  fprintf(outf,"_a%d = A%d[%d]; \\\n",m,m,k);
	else 
	  fprintf(outf,"_a%d = A[%d]; \\\n",m,m);

	nspace();
	for (n=0;n<N;n++)
	  fprintf(outf,"c%d%d += _a%d*_b%d; ",m,n,m,n);
	fprintf(outf,"\\\n");
      }
      if (opA == Transpose) {
	nspace();
	fprintf(outf,"A += Astride; \\\n");
      }
    }

  }
  unnest();
  nspace();
  fprintf(outf,"}\n");

  fprintf(outf,"\n\n");

}
  
void
gen_lsz_macro(const int M, const int N,
	      int fp_dtype_num)
{
  const char *fp_dtype = fp_types[fp_dtype_num];
  int m,n;

  if (!UseZeroScaleAccumStore) {
    nspace();
    fprintf(outf,"#define LOAD%dx%d(",
	    M,N);
    for (m=0;m<M;m++)
      for (n=0;n<N;n++)
	fprintf(outf,"c%d%d,",m,n);
    fprintf(outf,"C,Cstride) \\\n");
    
    nspace(); fprintf(outf,"{\\\n");
    
    nest();
    nspace(); fprintf(outf,"%s * _cp = C; \\\n",fp_dtype);
    
    for (m=0;m<M;m++) {
      nspace();
      for (n=0;n<N;n++)
	fprintf(outf,"c%d%d = _cp[%d]; ",m,n,n);
      fprintf(outf,"\\\n");
      if (m < (M-1)) {
	nspace(); fprintf(outf,"_cp += Cstride; \\\n");
      }
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n\n");
    
    
    
    nspace();
    fprintf(outf,"#define STORE%dx%d(",
	    M,N);
    for (m=0;m<M;m++)
      for (n=0;n<N;n++)
	fprintf(outf,"c%d%d,",m,n);
    fprintf(outf,"C,Cstride) \\\n");
    
    nspace(); fprintf(outf,"{\\\n");
    
    nest();
    nspace(); fprintf(outf,"%s *_cp = C; \\\n",fp_dtype);
    
    for (m=0;m<M;m++) {
      nspace();
      for (n=0;n<N;n++)
	fprintf(outf,"_cp[%d] = c%d%d; ",n,m,n);
      fprintf(outf,"\\\n");
      if (m < (M-1)) {
	nspace(); fprintf(outf,"_cp += Cstride; \\\n");
      }
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n\n");
  } else {
    nspace();
    fprintf(outf,"#define ZERO%dx%d(",
	    M,N);
    for (m=0;m<M;m++)
      for (n=0;n<N;n++) {
	if (n == N-1 && m == M-1)
	  fprintf(outf,"c%d%d",m,n);
	else 
	  fprintf(outf,"c%d%d,",m,n);
      }
    fprintf(outf,") \\\n");
    
    nspace(); fprintf(outf,"{\\\n");
    nest();
    
    for (m=0;m<M;m++) {
      nspace();
      for (n=0;n<N;n++)
	fprintf(outf,"c%d%d = 0.0; ",m,n);
      fprintf(outf,"\\\n");
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n\n");
    
    nspace();
    fprintf(outf,"#define SCALE_ACCUMULATE_STORE%dx%d(",
	    M,N);
    for (m=0;m<M;m++)
      for (n=0;n<N;n++)
	fprintf(outf,"c%d%d,",m,n);
    fprintf(outf,"C,Cstride) \\\n");
    
    nspace(); fprintf(outf,"{\\\n");
    
    nest();
    nspace(); fprintf(outf,"%s *_cp = C; \\\n",fp_dtype);
    
    for (m=0;m<M;m++) {
      nspace();
      for (n=0;n<N;n++) {
	if (alpha == Arb)
	  fprintf(outf,"_cp[%d] += alpha*c%d%d; ",n,m,n);
	else 
	  fprintf(outf,"_cp[%d] += %s*c%d%d; ",n,alpha_val,m,n);
      }
      fprintf(outf,"\\\n");
      if (m < (M-1)) {
	nspace(); fprintf(outf,"_cp += Cstride; \\\n");
      }
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n\n");
  }
}

void
generate_macros() 
{

  int m,k,n;
  int loc_fp_dtype_num;

  if (sepaccum) {
    assert ( cache_blocking_entries > 1 );
    loc_fp_dtype_num = fp_atype_num;
  } else 
    loc_fp_dtype_num = fp_dtype_num;

  if (do_fringes) {
    m = 1;
    while (1) {
      n = 1;
      while (1) {
	gen_lsz_macro(m,n,loc_fp_dtype_num);
	k = 1;
	while (1) { 
	  gen_mm_macro(m,k,n,loc_fp_dtype_num); 
	  if (k == cache_blocking[0].K)
	    break;
	  if ((k << 1) > cache_blocking[0].K) 
	    k = cache_blocking[0].K;
	  else 
	    k <<= 1;
	}
	if (n == cache_blocking[0].N)
	  break;
	if ((n << 1) > cache_blocking[0].N) 
	  n = cache_blocking[0].N;
	else 
	  n <<= 1;
      }
      if (m == cache_blocking[0].M)
	break;
      if ((m << 1) > cache_blocking[0].M) 
	m = cache_blocking[0].M;
      else 
	m <<= 1;
    }
  } else {
    m = cache_blocking[0].M;
    n = cache_blocking[0].N;
    k = cache_blocking[0].K;
    gen_lsz_macro(m,n,loc_fp_dtype_num);
    gen_mm_macro(m,k,n,loc_fp_dtype_num);
  }
}


void
generate_beta_code()
{
  if (beta == One)
    return;  /* no scaling necessary */

  if (beta == Arb || beta == Fix) {
    if (beta == Fix || !opt_beta) {
      nspace(); fprintf(outf,"{\n");
      nest(); /* 1 */
      nspace(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      nspace(); fprintf(outf,"cpre = C + M*Cstride;\n");
      nspace();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nest(); /* 2 */
      nspace(); fprintf(outf,"cpe = cprb + N;\n");
      nspace(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nest(); /* 3 */
      nspace(); fprintf(outf,"*cp *= (%s);\n",(beta==Arb?"beta":beta_val));
      unnest(); /* 3 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 2 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 1 */
      nspace(); fprintf(outf,"}\n");
    } else { /* beta == Arb && opt_beta */
      nspace(); fprintf(outf,"if (beta == 0.0) {\n");
      nest(); /* 1 */
      nspace(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      nspace(); fprintf(outf,"cpre = C + M*Cstride;\n");
      nspace();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nest(); /* 2 */
      nspace(); fprintf(outf,"cpe = cprb + N;\n");
      nspace(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nest(); /* 3 */
      nspace(); fprintf(outf,"*cp = 0.0;\n");
      unnest(); /* 3 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 2 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 1 */
      nspace(); fprintf(outf,"} else if (beta != 1.0) {\n");
      nest(); /* 1 */
      nspace(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      nspace(); fprintf(outf,"cpre = C + M*Cstride;\n");
      nspace();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nest(); /* 2 */
      nspace(); fprintf(outf,"cpe = cprb + N;\n");
      nspace(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nest(); /* 3 */
      nspace(); fprintf(outf,"*cp *= (beta);\n");
      unnest(); /* 3 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 2 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 1 */
      nspace(); fprintf(outf,"}\n");
    }
  } else if (beta == Zero) {
      nspace(); fprintf(outf,"{\n");
      nest(); /* 1 */
      nspace(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
      nspace(); fprintf(outf,"cpre = C + M*Cstride;\n");
      nspace();
      fprintf(outf,"for (cprb = C; cprb != cpre; cprb += Cstride) {\n");
      nest(); /* 2 */
      nspace(); fprintf(outf,"cpe = cprb + N;\n");
      nspace(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
      nest(); /* 3 */
      nspace(); fprintf(outf,"*cp = 0.0;\n");
      unnest(); /* 3 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 2 */
      nspace(); fprintf(outf,"}\n");    
      unnest(); /* 1 */
      nspace(); fprintf(outf,"}\n");
  }
}

/*
**
**  Core (register block) code generation.
**
*/



void
generate_core_rest_k(const int cb_entry,
		     const int rbm,const int rbk,const int rbn,
		     int fp_dtype_num)
{
  int bitk;
  int m,n;

  if (rbk>1) {
    bitk = (pow_of_2(rbk) ? (rbk >> 1) : highbit(rbk)); 
    for (;bitk>0;bitk>>=1) {

      nspace(); fprintf(outf,"if (k_marg_el & 0x%X) {\n",bitk);
      nest();

      nspace(); 
      generate_macro_name(rbm,bitk,rbn,fp_dtype_num);
      fprintf(outf,"(");
      for (m=0;m<rbm;m++)
	for (n=0;n<rbn;n++)
	  fprintf(outf,"c%d%d,",m,n);
      if (opA == Normal) {
	for (m=0;m<rbm;m++)
	  fprintf(outf,"ap%d_%d,",cb_entry,m);
      } else {
	fprintf(outf,"ap%d,Astride,",cb_entry);
      }
      if (opB == Normal) { 
	fprintf(outf,"bp%d,Bstride);\n",cb_entry);
      } else {
	for (n=0;n<rbn;n++)
	  fprintf(outf,"bp%d_%d%s",cb_entry,n,((n==(rbn-1))?"":","));
	fprintf(outf,");\n");
      }

      if (bitk > 1) {
	if (opA == Normal) {
	  nspace();
	  for (m=0;m<rbm;m++)
	    fprintf(outf,"ap%d_%d+=%d;",cb_entry,m,bitk);
	  fprintf(outf,"\n");
	}
	if (opB == Transpose) {
	  nspace();
	  for (n=0;n<rbn;n++)
	    fprintf(outf,"bp%d_%d+=%d;",cb_entry,n,bitk);
	  fprintf(outf,"\n");
	}
      }

      unnest();
      nspace(); fprintf(outf,"}\n");
    }
  }
}


void
generate_core_block_k(const int cb_entry,
		      const int rbm,const int rbk,const int rbn,
		      int fp_dtype_num)
{
  int m,n;

  if (opA == Normal) {
    nspace(); fprintf(outf,"ap%d_0 = a%d;\n",cb_entry,cb_entry);
    for (m=1;m<rbm;m++) {
      nspace(); fprintf(outf,"ap%d_%d = ap%d_%d + Astride;\n",cb_entry,m,cb_entry,m-1);
    }
  } else {
    nspace();
    fprintf(outf,"ap%d=a%d;\n",cb_entry,cb_entry);
  }

  if (opB == Transpose) {
    nspace(); fprintf(outf,"bp%d_0 = b%d;\n",cb_entry,cb_entry);
    for (n=1;n<rbn;n++) {
      nspace(); fprintf(outf,"bp%d_%d = bp%d_%d + Bstride;\n",cb_entry,n,cb_entry,n-1);
    }
  } else {
    nspace();
    fprintf(outf,"bp%d=b%d;\n",cb_entry,cb_entry);
  }

  nspace();
  if (!UseZeroScaleAccumStore)
    fprintf(outf,"LOAD%dx%d(",
		      rbm,rbn);
  else 
    fprintf(outf,"ZERO%dx%d(",
		      rbm,rbn);
  for (m=0;m<rbm;m++)
    for (n=0;n<rbn;n++)
      fprintf(outf,"c%d%d%s",m,n,((m==(rbm-1))&&(n==(rbn-1))?"":","));
  if (!UseZeroScaleAccumStore)
    fprintf(outf,",cp%d,Cstride",cb_entry);
  fprintf(outf,");\n");


  nspace();
  fprintf(outf,"for (; ");
  fprintf(outf,"ap%d%s!=ap%d_endp; ",
	  cb_entry,
	  (opA == Normal ? "_0" : ""),
	  cb_entry);
  if (opA == Normal) {
    for (m=0;m<rbm;m++)
      fprintf(outf,"ap%d_%d+=%d%s",cb_entry,m,rbk,
	      (m<(rbm-1))?",":"");
  } else {
    /* do nothing */
  }
  if (opB == Normal) {
    /* do nothing */
  } else {
    if (opA == Normal) 
      fprintf(outf,",");
    for (n=0;n<rbn;n++)
      fprintf(outf,"bp%d_%d+=%d%s",cb_entry,n,rbk,
	      (n<(rbn-1))?",":"");
  }
  fprintf(outf,") {\n");
  nest();

  nspace(); 
  generate_macro_name(rbm,rbk,rbn,fp_dtype_num);
  fprintf(outf,"(");
  for (m=0;m<rbm;m++)
    for (n=0;n<rbn;n++)
      fprintf(outf,"c%d%d,",m,n);
  if (opA == Normal) {
    for (m=0;m<rbm;m++)
      fprintf(outf,"ap%d_%d,",cb_entry,m);
  } else {
    fprintf(outf,"ap%d,Astride,",cb_entry);
  }
  if (opB == Normal) {
    fprintf(outf,"bp%d,Bstride);\n",cb_entry);
  } else {
    for (n=0;n<rbn;n++)
      fprintf(outf,"bp%d_%d%s",cb_entry,n,((n==(rbn-1))?"":","));
    fprintf(outf,");\n");
  }

  unnest();
  nspace(); fprintf(outf,"}\n");
}


void
generate_core_end_k(const int cb_entry,const int rbm,const int rbk,const int rbn)
{
  int m,n;
  nspace(); 
  if (!UseZeroScaleAccumStore)
    fprintf(outf,"STORE%dx%d(",
	    rbm,rbn);
  else 
    fprintf(outf,"SCALE_ACCUMULATE_STORE%dx%d(",
	    rbm,rbn);    
  for (m=0;m<rbm;m++)
    for (n=0;n<rbn;n++)
      fprintf(outf,"c%d%d,",m,n);
  fprintf(outf,"cp%d,Cstride);\n",cb_entry);
}

void
generate_core_rest_n(const int cb_entry,
		     const int rbm,const int rbk,const int rbn,
		     int fp_dtype_num
		     )
{
  int bitn;
  if (rbn > 1 ) {
    bitn = (pow_of_2(rbn) ? (rbn >> 1) : highbit(rbn)); 
    for (;bitn>0;bitn>>=1) {
      nspace(); fprintf(outf,"if (n_marg_el & 0x%X) {\n",bitn);
      nest();


      generate_core_block_k(cb_entry,rbm,rbk,bitn,fp_dtype_num);
      generate_core_rest_k(cb_entry,rbm,rbk,bitn,fp_dtype_num);
      generate_core_end_k(cb_entry,rbm,rbk,bitn);
	
      if (bitn > 1) {
	if (opB == Normal) {
	  nspace(); fprintf(outf,"b%d += %d;\n",cb_entry,bitn);
	} else {
	  nspace(); fprintf(outf,"b%d += Bstride*%d;\n",cb_entry,bitn);
	}
	nspace(); fprintf(outf,"cp%d += %d;\n",cb_entry,bitn);
      }

      unnest();
      nspace(); fprintf(outf,"}\n");
    }
  }
}

void
generate_core_n_loop (const bool arbitrary,
		      const int cb_entry,
		      const int rbn,
		      const char *fp_dtype)
{
  /* generate the N loop */

  nspace(); fprintf(outf,"%s* const cp%d_endp = ",fp_dtype,cb_entry);
  if (arbitrary)
    fprintf(outf,"c%d + n_norm",cb_entry);
  else 
    fprintf(outf,"c%d + %d",cb_entry,cache_blocking[cb_entry+1].cumN);
  fprintf(outf,";\n");

  nspace();
  fprintf(outf,"for (b%d=B,cp%d=c%d; cp%d!=cp%d_endp; ",
	    cb_entry,cb_entry,cb_entry,
	    cb_entry,cb_entry);
  fprintf(outf,"b%d+=",
	  cb_entry);
  if (opB == Normal) {
    fprintf(outf,"%d,",rbn);
  } else {
    fprintf(outf,"B_sbs_stride,");
  }
  fprintf(outf,"cp%d+=%d) {\n",
	  cb_entry,rbn);
  nest();
}

void
generate_core_m_loop(const bool arbitrary,
		     const int cb_entry,
		     const int rbm)
{
  /* generate M loop */
  nspace();
  fprintf(outf,"for (c%d=C,a%d=A; c%d!= c%d_endp; c%d+=C_sbs_stride,",
	    cb_entry,cb_entry,cb_entry,cb_entry,cb_entry);
  if (opA == Normal) {
    fprintf(outf,"a%d+=A_sbs_stride",
	    cb_entry);
  } else {
    fprintf(outf,"a%d+=%d",
	    cb_entry,rbm);
  }
  fprintf(outf,") {\n");
  nest();
}


/* Generate the core register-blocked matrix-matrix code where all register
** blocking is unrolled via calls to previously defined macros. 
*/
void
generate_core_code(bool top_level, 
		   routineKind kind,
		   int fp_dtype_num)
{
  int m,n;
  int bitm;
  const int cb_entry = 0;
  /* load these here just for ease of typing and sight */
  const int rbm = cache_blocking[cb_entry].M;
  const int rbn = cache_blocking[cb_entry].N;
  const int rbk = cache_blocking[cb_entry].K;
  const char*fp_dtype = fp_types[fp_dtype_num];

  if (top_level)
    assert(kind == Arb_all);

  nspace();
  fprintf(outf,"const %s *a%d,*b%d;\n",fp_type,cb_entry,cb_entry);
  nspace();
  fprintf(outf,"%s *c%d;\n",fp_dtype,cb_entry);

  nspace();
  fprintf(outf,"const %s ",fp_type);
  if (opA == Normal) {
    for (m=0;m<rbm;m++) 
      fprintf(outf,"*ap%d_%d%s",cb_entry,m,((m==(rbm-1))?";":","));
    fprintf(outf,"\n");
  } else {
    fprintf(outf,"*ap%d;\n",cb_entry);
  }
  nspace();
  fprintf(outf,"const %s ",fp_type);
  if (opB == Normal) {
    fprintf(outf,"*bp%d;\n",cb_entry);
  } else {
    for (n=0;n<rbn;n++) 
      fprintf(outf,"*bp%d_%d%s",cb_entry,n,((n==(rbn-1))?";":","));
    fprintf(outf,"\n");
  }

  nspace();
  fprintf(outf,"%s *cp%d;\n",fp_dtype,cb_entry);

  if (opA == Normal) {
    nspace(); fprintf(outf,"const int A_sbs_stride = Astride*%d;\n",rbm);
  }
  if (opB == Transpose) {
    nspace(); fprintf(outf,"const int B_sbs_stride = Bstride*%d;\n",rbn);
  }
  nspace(); fprintf(outf,"const int C_sbs_stride = Cstride*%d;\n",rbm);

  if (kind != Fixed) {
    if (kind == Arb_k || kind == Arb_all) {
      nspace(); fprintf(outf,"const int k_marg_el = ");
      if (pow_of_2(rbk))
	fprintf(outf,"K & %d;\n",rbk - 1);
      else 
	fprintf(outf,"K %% %d;\n",rbk);
      nspace(); 
      if (opA == Normal)
	fprintf(outf,"const int k_norm = K - k_marg_el;\n");
      else 
	fprintf(outf,"const int k_norm = (K - k_marg_el)*Astride;\n");
    }
    if (kind == Arb_all) {
      nspace(); fprintf(outf,"const int m_marg_el = ");
      if (pow_of_2(rbm))
	fprintf(outf,"M & %d;\n",rbm - 1);
      else 
	fprintf(outf,"M %% %d;\n",rbm);
      nspace(); fprintf(outf,"const int m_norm = M - m_marg_el;\n");
      nspace(); fprintf(outf,"const int n_marg_el = ");
      if (pow_of_2(rbn))
	fprintf(outf,"N & %d;\n",rbn - 1);
      else 
	fprintf(outf,"N %% %d;\n",rbn);
      nspace(); fprintf(outf,"const int n_norm = N - n_marg_el;\n");
    }
  } else {
    assert ( !top_level );
    if (opA == Transpose) {
      nspace();
      fprintf(outf,"const int k_norm = %d*Astride;\n",
	      cache_blocking[cb_entry+1].cumK);
    }
  }

  if (top_level || kind == Arb_all) {
    nspace();
    fprintf(outf,"%s *const c%d_endp = C+m_norm*Cstride;\n",fp_dtype,cb_entry);
  } else {
    nspace();
    fprintf(outf,"%s *const c%d_endp = C+%d*Cstride;\n",
	    fp_dtype,cb_entry,cache_blocking[cb_entry+1].cumM);
  }


  nspace(); fprintf(outf,"register %s ",fp_atype);
  for (m=0;m<rbm;m++)
    for (n=0;n<rbn;n++)
      fprintf(outf,"c%d%d%s",m,n,(m==(rbm-1)&&(n==(rbn-1))?"":","));
  fprintf(outf,";\n");


  if (top_level && !sepaccum)
    generate_beta_code();
  if (top_level && alpha == Arb) {
    nspace(); fprintf(outf,"if (alpha == 0.0) {\n");
    nest(); /* 1 */
    if (!sepaccum) {
      nspace(); fprintf(outf,"return;\n");
    } else {
      /* we still may need to scale beta */
      generate_beta_code();
      nspace(); fprintf(outf,"return;\n");
    }
    unnest(); /* 1 */
    nspace(); fprintf(outf,"}\n");
  }

  generate_core_m_loop(top_level || kind == Arb_all,cb_entry,rbm);

  nspace(); fprintf(outf,"const %s* const ap%d_endp = ",fp_type,cb_entry);
  if (top_level || kind == Arb_all || kind == Arb_k || opA == Transpose) 
    fprintf(outf,"a%d + k_norm",cb_entry);
  else 
    fprintf(outf,"a%d + %d",cb_entry,cache_blocking[cb_entry+1].cumK);
  fprintf(outf,";\n");
  
  generate_core_n_loop(top_level || kind == Arb_all,cb_entry,rbn,fp_dtype);

  generate_core_block_k(cb_entry,rbm,rbk,rbn,fp_dtype_num);
  if ((rbk>1) && do_fringes && (top_level || kind == Arb_all || kind == Arb_k))
    { 
      generate_core_rest_k(cb_entry,rbm,rbk,rbn,fp_dtype_num);
    }
  generate_core_end_k(cb_entry,rbm,rbk,rbn);

  unnest();
  nspace(); fprintf(outf,"}\n");

  unnest();
  nspace(); fprintf(outf,"}\n");

  if (do_fringes && (top_level || kind == Arb_all)) {
    /* do n and m fringe */

    if (rbn > 1) {
      /* do n fringe */

      /* generate M loop */
      generate_core_m_loop(False,cb_entry,rbm);
      
      nspace();
      fprintf(outf,"const %s* const ap%d_endp = a%d + k_norm;\n",
	      fp_type,cb_entry,cb_entry);

      if (opB == Normal)  {
	nspace(); fprintf(outf,"b%d = B+n_norm;\n",cb_entry);
      } else {
	nspace(); fprintf(outf,"b%d = B+n_norm*Bstride;\n",cb_entry);
      }
      nspace(); fprintf(outf,"cp%d = c%d+n_norm;\n",cb_entry,cb_entry);

      generate_core_rest_n(cb_entry,rbm,rbk,rbn,fp_dtype_num);

      unnest();
      nspace(); fprintf(outf,"}\n");

    }

    if (rbm > 1) {
      /* do m fringe */
      bitm = (pow_of_2(rbm) ? (rbm >> 1) : highbit(rbm)); 
      for (;bitm>0;bitm>>=1) {

	nspace(); fprintf(outf,"if (m_marg_el & 0x%X) {\n",bitm);
	nest();

        nspace();
        fprintf(outf,"const %s* const ap%d_endp = a%d + k_norm;\n",
		fp_type,cb_entry,cb_entry);

	generate_core_n_loop(True,cb_entry,rbn,fp_dtype);

	generate_core_block_k(cb_entry,bitm,rbk,rbn,fp_dtype_num);
	generate_core_rest_k(cb_entry,bitm,rbk,rbn,fp_dtype_num);
	generate_core_end_k(cb_entry,bitm,rbk,rbn);
	
	unnest();
	nspace(); fprintf(outf,"}\n");

	generate_core_rest_n(cb_entry,bitm,rbk,rbn,fp_dtype_num);

	if (bitm > 1) {
	  nspace(); fprintf(outf,"c0 += Cstride*%d;\n",bitm);
	  if (opA == Normal)  {
	    nspace(); fprintf(outf,"a0 += Astride*%d;\n",bitm);
	  } else {
	    nspace(); fprintf(outf,"a0 += %d;\n",bitm);
	  }
	}

	unnest();
	nspace(); fprintf(outf,"}\n");	
      }
    }
  }
}


/*
**
**  Level (L1 or higher cache block) code generation.
**
*/


void
generate_level_zeroaccum(char *len)
{
  nspace(); fprintf(outf,"{\n");
  nest(); /* 1 */
  nspace(); fprintf(outf,"%s*C_accum_p,*C_accum_endp;\n",fp_atype);
  nspace(); fprintf(outf,"C_accum_p = C_accum;\n");
  nspace(); fprintf(outf,"C_accum_endp = &C_accum[%s];\n",len);
  nspace(); fprintf(outf,"while (C_accum_p != C_accum_endp)\n");
  nest(); /* 2 */
  nspace(); fprintf(outf,"*C_accum_p++ = 0.0;\n");
  unnest(); /* 2 */
  unnest(); /* 1*/
  nspace(); fprintf(outf,"}\n");
}


void
generate_level_copyaccumback_loops(char *mlen,
				   char *nlen,
				   char *dest,
				   char *stride,
				   char *store_str, /* store or accumulate */
				   char *alphastr)
{
  nspace(); fprintf(outf,"for (cprb = %s; cprb != cpre; cprb += %s) {\n",
		    dest,stride);
  nest();
  nspace(); fprintf(outf,"cpe = cprb + %s;\n",nlen);
  nspace(); fprintf(outf,"for (cp = cprb; cp != cpe; cp++) {\n");
  nest();

  nspace();
  fprintf(outf,"*cp %s %s(*C_accum_p++);\n",store_str,alphastr);


  unnest();
  nspace(); fprintf(outf,"}\n");  
  unnest();
  nspace(); fprintf(outf,"}\n");  

}

void
generate_level_copyaccumback_setup(char *mlen,
				   char *nlen,
				   char *dest,
				   char *stride)
{
  nspace(); fprintf(outf,"%s *cprb,*cpre,*cp,*cpe;\n",fp_dtype);  
  nspace(); fprintf(outf,"%s *C_accum_p;\n",fp_atype);
  nspace(); fprintf(outf,"C_accum_p = C_accum;\n");
  nspace(); fprintf(outf,"cpre = %s + %s*%s;\n",dest,mlen,stride);
}

void
generate_level_copyaccumback(char *mlen,
			     char *nlen,
			     char *dest,
			     char *stride)
{
  char store_str[BUFSIZ];
  char alpha_str[BUFSIZ];

  if (beta == Arb) {
    if (alpha == Arb) {
      if (opt_alpha) {
	if (opt_beta) {
	  nspace(); fprintf(outf,"{\n");
	  nest(); /* 1 */
	  generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
	
	  nspace(); fprintf(outf,"if (alpha != 1.0) {\n");
	  nest(); /* 2 */

	  nspace(); fprintf(outf,"if (beta == 0.0) {\n");
	  nest(); /* 3 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "=",
					     "alpha*");
	  unnest(); /* 3 */
	  nspace(); fprintf(outf,"} else if (beta == 1.0) {\n");
	  nest(); /* 3 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "+=",
					     "alpha*");
	  unnest(); /* 3 */
	  nspace(); fprintf(outf,"} else {\n");
	  nest(); /* 3 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "= beta*(*cp) + ",
					     "alpha*");
	  unnest(); /* 3 */
	  nspace(); fprintf(outf,"}\n");


	  unnest(); /* 2 */
	  nspace(); fprintf(outf,"} else {\n");
	  nest(); /* 2 */

	  nspace(); fprintf(outf,"if (beta == 0.0) {\n");
	  nest(); /* 3 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "=",
					     "");
	  unnest(); /* 3 */
	  nspace(); fprintf(outf,"} else if (beta == 1.0) {\n");
	  nest(); /* 3 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "+=",
					     "");
	  unnest(); /* 3 */
	  nspace(); fprintf(outf,"} else {\n");
	  nest(); /* 3 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "= beta*(*cp) + ",
					     "");
	  unnest(); /* 3 */
	  nspace(); fprintf(outf,"}\n");

	  unnest(); /* 2 */
	  nspace(); fprintf(outf,"}\n");
	  unnest(); /* 1 */
	  nspace(); fprintf(outf,"}\n");
	} else /* !opt_beta */ {
	  nspace(); fprintf(outf,"{\n");
	  nest(); /* 1 */
	  generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
	  
	  nspace(); fprintf(outf,"if (alpha != 1.0) {\n");
	  nest(); /* 2 */
	  
	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "= beta*(*cp) +",
					     "alpha*");
	  unnest(); /* 2 */
	  nspace(); fprintf(outf,"} else {\n");
	  nest(); /* 2 */
	  
	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "= beta*(*cp) +",
					     "");
	  unnest(); /* 2 */
	  nspace(); fprintf(outf,"}\n");
	  unnest(); /* 1 */
	  nspace(); fprintf(outf,"}\n");
	}
      } else /* !opt_alpha */ {
	if (opt_beta) {
	  nspace(); fprintf(outf,"{\n");
	  nest(); /* 1 */
	  generate_level_copyaccumback_setup(mlen,nlen,dest,stride);

	  nspace(); fprintf(outf,"if (beta == 0.0) {\n");
	  nest(); /* 2 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "=",
					     "alpha*");
	  unnest(); /* 2 */
	  nspace(); fprintf(outf,"} else if (beta == 1.0) {\n");
	  nest(); /* 2 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "+=",
					     "alpha*");
	  unnest(); /* 2 */
	  nspace(); fprintf(outf,"} else {\n");
	  nest(); /* 2 */

	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "= beta*(*cp) + ",
					     "alpha*");
	  unnest(); /* 2 */
	  nspace(); fprintf(outf,"}\n");

	  unnest(); /* 1 */
	  nspace(); fprintf(outf,"}\n");
	} else /* !opt_beta */ {
	  nspace(); fprintf(outf,"{\n");
	  nest(); /* 1 */
	  generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
	  generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					     "= beta*(*cp) +",
					     "alpha*");
	  unnest(); /* 1 */
	  nspace(); fprintf(outf,"}\n");
	}
      }
    } else /* alpha != Arb */ {
      if (opt_beta) {
	if (alpha == One)
	  sprintf(alpha_str,"%s","");
	else 
	  sprintf(alpha_str,"(%s)*",alpha_val);

	nspace(); fprintf(outf,"{\n");
	nest(); /* 1 */
	generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
	
	nspace(); fprintf(outf,"if (beta == 0.0) {\n");
	nest(); /* 2 */
	
	generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					   "=",
					   alpha_str);
	unnest(); /* 2 */
	nspace(); fprintf(outf,"} else if (beta == 1.0) {\n");
	nest(); /* 2 */
	
	generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					   "+=",
					   alpha_str);
	unnest(); /* 2 */
	nspace(); fprintf(outf,"} else {\n");
	nest(); /* 2 */
	
	generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					   "= beta*(*cp) + ",
					   alpha_str);
	unnest(); /* 2 */
	nspace(); fprintf(outf,"}\n");
	
	unnest(); /* 1 */
	nspace(); fprintf(outf,"}\n");
      } else /* !opt_beta */ {
	if (alpha == One)
	  sprintf(alpha_str,"%s","");
	else 
	  sprintf(alpha_str,"(%s)*",alpha_val);
	nspace(); fprintf(outf,"{\n");
	nest(); /* 1 */
	generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
	generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					   "= beta*(*cp) +",
					   alpha_str);
	unnest(); /* 1 */
	nspace(); fprintf(outf,"}\n");
      }
    }
  } else /* beta != Arb */ {
    if (alpha == Arb) {
      if (opt_alpha) {
	nspace(); fprintf(outf,"{\n");
	nest(); /* 1 */
	generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
	if (beta == One)
	  sprintf(store_str,"+=");
	else if (beta == Zero)
	  sprintf(store_str,"=");
	else /* beta == Fix */
	  sprintf(store_str,"= (%s)*(*cp) + ",beta_val);	  
	
	nspace(); fprintf(outf,"if (alpha != 1.0) {\n");
	nest(); /* 2 */

	generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					   store_str,
   					   "alpha*");
  	unnest(); /* 2 */
	nspace(); fprintf(outf,"} else {\n");
	nest(); /* 2 */

	generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					   store_str,
					   "");
	unnest(); /* 2 */
	nspace(); fprintf(outf,"}\n");
	unnest(); /* 1 */
	nspace(); fprintf(outf,"}\n");
      } else /* !opt_alpha */ {
	nspace(); fprintf(outf,"{\n");
	nest(); /* 1 */
	generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
	if (beta == One)
	  sprintf(store_str,"+=");
	else if (beta == Zero)
	  sprintf(store_str,"=");
	else /* beta == Fix */
	  sprintf(store_str,"= (%s)*(*cp) + ",beta_val);	  
	generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					   store_str,
					   "alpha*");
	unnest(); /* 1 */
	nspace(); fprintf(outf,"}\n");
      }
    } else /* alpha != Arb */ {
      nspace(); fprintf(outf,"{\n");
      nest(); /* 1 */
      generate_level_copyaccumback_setup(mlen,nlen,dest,stride);
      if (beta == One)
	sprintf(store_str,"+=");
      else if (beta == Zero)
	sprintf(store_str,"=");
      else /* beta == Fix */
	sprintf(store_str,"= (%s)*(*cp) + ",beta_val);	  
      if (alpha == One)
	sprintf(alpha_str,"%s","");
      else 
	sprintf(alpha_str,"(%s)*",alpha_val);
      generate_level_copyaccumback_loops(mlen,nlen,dest,stride,
					 store_str,
					 alpha_str);
      unnest(); /* 1 */
      nspace(); fprintf(outf,"}\n");
    }
  }
}

void
generate_level_m_loop(bool arbitrary,
		      int cb_level)
{
  /* M loop */
  nspace();
  if (arbitrary) {
    fprintf(outf,"for (m%d=0; m%d<=M-%d; m%d+=%d) {\n",
	    cb_level,cb_level,cache_blocking[cb_level-1].cumM,
	    cb_level,cache_blocking[cb_level-1].cumM);
  } else {
    fprintf(outf,"for (m%d=0; m%d<%d; m%d+=%d) {\n",
	    cb_level,cb_level,cache_blocking[cb_level].cumM,
	    cb_level,
	    cache_blocking[cb_level-1].cumM);
  }
  nest();
  nspace(); fprintf(outf,"c%d = C + m%d*Cstride;\n",cb_level,cb_level);
  if (opA == Normal) {
    nspace(); fprintf(outf,"a%d = A + m%d*Astride;\n",cb_level,cb_level);
  } else {
    nspace(); fprintf(outf,"a%d = A + m%d;\n",cb_level,cb_level);
  }
}

void
generate_level_n_loop(bool arbitrary,
		      int cb_level)
{
  /* N loop */
  nspace();
  fprintf(outf,"for (n%d=0,b%d=B,cp%d=c%d; ",
	    cb_level,cb_level,cb_level,cb_level);
  if (arbitrary) {
    fprintf(outf,"n%d<=N-%d; ",
	    cb_level,cache_blocking[cb_level-1].cumN);
  } else {
    fprintf(outf,"n%d<%d; ",
	    cb_level,cache_blocking[cb_level].cumN);
  }
  fprintf(outf,"n%d+=%d,",cb_level,cache_blocking[cb_level-1].cumN);
  if (opB == Normal) {
    fprintf(outf,"b%d+=%d,",cb_level,cache_blocking[cb_level-1].cumN);
  } else {
    fprintf(outf,"b%d+=%d*Bstride,",cb_level,cache_blocking[cb_level-1].cumN);
  }
  fprintf(outf,"cp%d+=%d) {\n",
	  cb_level,cache_blocking[cb_level-1].cumN);
  nest();
}



void
generate_level_k_loop(bool arbitrary,
		      int cb_level)
{
  /* K loop */
  nspace();
  fprintf(outf,
	  "for (k%d=0,bp%d=b%d,ap%d=a%d; ",
	  cb_level,cb_level,cb_level,cb_level,cb_level);
  if (arbitrary) {
    fprintf(outf,
	    "k%d<=K-%d; ",
	    cb_level,cache_blocking[cb_level-1].cumK);
  } else {
    fprintf(outf,"k%d<%d; ",
	    cb_level,cache_blocking[cb_level].cumK);
  }
  fprintf(outf,"k%d+=%d,",
	  cb_level,cache_blocking[cb_level-1].cumK);
  if (opB == Normal) {
    fprintf(outf,
	    "bp%d+=%d*Bstride,",
	    cb_level,cache_blocking[cb_level-1].cumK);
  } else {
    fprintf(outf,
	    "bp%d+=%d,",
	    cb_level,cache_blocking[cb_level-1].cumK);
  }
  if (opA == Normal) {
    fprintf(outf,
	    "ap%d+=%d) {\n",
	    cb_level,cache_blocking[cb_level-1].cumK);
  } else {
    fprintf(outf,
	    "ap%d+=%d*Astride) {\n",
	    cb_level,cache_blocking[cb_level-1].cumK);
  }
  nest();
}



void
generate_level_code(bool top_level,  /* True if code for first called routine */
		    int cb_level,     /* The current cache blocking level. */
		    routineKind kind,  /* Kind of routine */
		    int fp_dtype_num   /* the current destination type */
		    )
{
  const char *fp_dtype = fp_types[fp_dtype_num];
  assert ( cb_level >= 2 );

  if (top_level)
    assert(kind == Arb_all);

  nspace();
  fprintf(outf,"/* Code for L%d-blocked routine. */\n",cb_level-1);

  nspace();
  fprintf(outf,"int m%d,k%d,n%d;\n",cb_level,cb_level,cb_level);

  nspace();
  fprintf(outf,"const %s *a%d,*b%d;\n",fp_type,cb_level,cb_level);
  nspace();
  fprintf(outf,"%s *c%d;\n",fp_dtype,cb_level);


  nspace();
  fprintf(outf,"const %s *ap%d,*bp%d;\n",fp_type,cb_level,cb_level);
  nspace();
  fprintf(outf,"%s *cp%d;\n",fp_dtype,cb_level);

  if (top_level && !sepaccum)
    generate_beta_code();
  if (top_level && alpha == Arb) {
    nspace(); fprintf(outf,"if (alpha == 0.0) {\n");
    nest(); /* 1 */
    if (!sepaccum) {
      nspace(); fprintf(outf,"return;\n");
    } else {
      /* we still may need to scale beta */
      generate_beta_code();
      nspace(); fprintf(outf,"return;\n");
    }
    unnest(); /* 1 */
    nspace(); fprintf(outf,"}\n");
  }


  if (kind == Arb_all && call_down ) {
    nspace();
    fprintf(outf,"if (M < %d && K < %d && N < %d) {\n",
	   cache_blocking[cb_level-1].cumM+1,
	   cache_blocking[cb_level-1].cumK+1,
	   cache_blocking[cb_level-1].cumN+1);
    nest();
    
    if (top_level && sepaccum) {
      generate_level_zeroaccum("M*N");
      nspace();
      fprintf(outf,
	      "%s_l%d_arb_all(M,K,N,A,B,C_accum,Astride,Bstride,N);\n",
	      routine_name,cb_level-1);
      generate_level_copyaccumback("M","N","C","Cstride");
    } else {
      nspace();
      fprintf(outf,
	      "%s_l%d_arb_all(M,K,N,A,B,C,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : ""));
    }

    nspace(); fprintf(outf,"return;\n");

    unnest();
    nspace(); fprintf(outf,"}\n");



  }

  generate_level_m_loop(top_level || kind == Arb_all,cb_level);

  generate_level_n_loop(top_level || kind == Arb_all,cb_level);  

  if (top_level && sepaccum) {
    char buff[BUFSIZ];
    sprintf(buff,"(%d*%d)",
	    cache_blocking[cb_level-1].cumM,
	    cache_blocking[cb_level-1].cumN);
    generate_level_zeroaccum(buff);
  }

  generate_level_k_loop(top_level || kind == Arb_k || kind == Arb_all,cb_level);
  if (top_level && sepaccum) {
    nspace();
    fprintf(outf,
	    "%s_l%d(ap%d,bp%d,C_accum,Astride,Bstride,%d);\n",
	    routine_name,cb_level-1,cb_level,cb_level,
	    cache_blocking[cb_level-1].cumN);

  } else {
    nspace();
    fprintf(outf,
	    "%s_l%d(ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	    routine_name,cb_level-1,cb_level,cb_level,cb_level,
	    (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
  }

  unnest();
  nspace(); fprintf(outf,"}\n");

  /* finish up K */
  if (top_level || kind == Arb_k || kind == Arb_all) {
    nspace();
    fprintf(outf,"if (k%d < K) {\n",cb_level);
    nest();

    if (top_level && sepaccum) {
      nspace();
      fprintf(outf,
	      "%s_l%d_arb_k(K-k%d,ap%d,bp%d,C_accum,Astride,Bstride,%d);\n",
	      routine_name,cb_level-1,cb_level,
	      cb_level,cb_level,
	      cache_blocking[cb_level-1].cumN);
    } else {
      nspace();
      fprintf(outf,
	      "%s_l%d_arb_k(K-k%d,ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,cb_level,
	      cb_level,cb_level,cb_level,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
    }

    unnest();
    nspace(); fprintf(outf,"}\n");
  }

  if (top_level && sepaccum) {
    char mlen[BUFSIZ],nlen[BUFSIZ],cname[BUFSIZ];
    sprintf(mlen,"%d",cache_blocking[cb_level-1].cumM);
    sprintf(nlen,"%d",cache_blocking[cb_level-1].cumN);
    sprintf(cname,"cp%d",cb_level);
    generate_level_copyaccumback(mlen,nlen,cname,"Cstride");
  }

  unnest(); 
  nspace(); fprintf(outf,"}\n");    

  /* finish up N */
  if (top_level || kind == Arb_all) {

    nspace(); fprintf(outf,"if (n%d < N) {\n",cb_level);
    nest();

    if (top_level && sepaccum) {
      char buff[BUFSIZ];
      sprintf(buff,"(%d*(N-n%d))",
	      cache_blocking[cb_level-1].cumM,
	      cb_level);
      generate_level_zeroaccum(buff);
    }

    generate_level_k_loop(True,cb_level);
    
    nspace();
    if (top_level && sepaccum) {
      fprintf(outf,
	      "%s_l%d_arb_all(%d,%d,N-n%d,ap%d,bp%d,C_accum,Astride,Bstride,(N-n%d));\n",
	      routine_name,cb_level-1,
	      cache_blocking[cb_level-1].cumM,
	      cache_blocking[cb_level-1].cumK,
	      cb_level,
	      cb_level,cb_level,
	      cb_level);
    } else {
      fprintf(outf,
	      "%s_l%d_arb_all(%d,%d,N-n%d,ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,
	      cache_blocking[cb_level-1].cumM,
	      cache_blocking[cb_level-1].cumK,
	      cb_level,
	      cb_level,cb_level,cb_level,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
    }

    
    unnest();
    nspace(); fprintf(outf,"}\n");


    nspace();
    fprintf(outf,"if (k%d < K) {\n",cb_level);
    nest();

    nspace();
    if (top_level && sepaccum) {
      fprintf(outf,
	      "%s_l%d_arb_all(%d,K-k%d,N-n%d,ap%d,bp%d,C_accum,Astride,Bstride,(N-n%d));\n",
	      routine_name,cb_level-1,
	      cache_blocking[cb_level-1].cumM,
	      cb_level,
	      cb_level,
	      cb_level,cb_level,
	      cb_level);
    } else {
      fprintf(outf,
	      "%s_l%d_arb_all(%d,K-k%d,N-n%d,ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,
	      cache_blocking[cb_level-1].cumM,
	      cb_level,
	      cb_level,
	      cb_level,cb_level,cb_level,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
    }

    unnest();
    nspace(); fprintf(outf,"}\n");

    if (top_level && sepaccum) {
      char mlen[BUFSIZ],nlen[BUFSIZ],cname[BUFSIZ];
      sprintf(mlen,"%d",cache_blocking[cb_level-1].cumM);
      sprintf(nlen,"(N-n%d)",cb_level);
      sprintf(cname,"cp%d",cb_level);
      generate_level_copyaccumback(mlen,nlen,cname,"Cstride");
    }

    unnest();
    nspace(); fprintf(outf,"}\n");
  }

  unnest();
  nspace(); fprintf(outf,"}\n");    

  /* finish up M */
  if (top_level || kind == Arb_all) {

    nspace(); fprintf(outf,"if (m%d < M) {\n",cb_level);
    nest();

    nspace(); fprintf(outf,"c%d = C + m%d*Cstride;\n",cb_level,cb_level);
    if (opA == Normal) {
      nspace(); fprintf(outf,"a%d = A + m%d*Astride;\n",cb_level,cb_level);
    } else {
      nspace(); fprintf(outf,"a%d = A + m%d;\n",cb_level,cb_level);
    }

    generate_level_n_loop(True,cb_level);

    if (top_level && sepaccum) {
      char buff[BUFSIZ];
      sprintf(buff,"((M-m%d)*%d)",
	      cb_level,
	      cache_blocking[cb_level-1].cumN);
      generate_level_zeroaccum(buff);
    }

    generate_level_k_loop(True,cb_level);

    nspace();
    if (top_level && sepaccum) {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,%d,%d,ap%d,bp%d,C_accum,Astride,Bstride,%d);\n",
	      routine_name,cb_level-1,
	      cb_level,
	      cache_blocking[cb_level-1].cumK,
	      cache_blocking[cb_level-1].cumN,
	      cb_level,cb_level,
	      cache_blocking[cb_level-1].cumN);
    } else {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,%d,%d,ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,
	      cb_level,
	      cache_blocking[cb_level-1].cumK,
	      cache_blocking[cb_level-1].cumN,
	      cb_level,cb_level,cb_level,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n");
    
    /* finish up K */
    nspace();
    fprintf(outf,"if (k%d < K) {\n",cb_level);
    
    nest();
    nspace();
    if (top_level && sepaccum) {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,K-k%d,%d,ap%d,bp%d,C_accum,Astride,Bstride,%d);\n",
	      routine_name,cb_level-1,
	      cb_level,cb_level,
	      cache_blocking[cb_level-1].cumN,
	      cb_level,cb_level,
	      cache_blocking[cb_level-1].cumN);
    } else {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,K-k%d,%d,ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,
	      cb_level,cb_level,
	      cache_blocking[cb_level-1].cumN,
	      cb_level,cb_level,cb_level,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n");


    if (top_level && sepaccum) {
      char mlen[BUFSIZ],nlen[BUFSIZ],cname[BUFSIZ];
      sprintf(mlen,"(M-m%d)",cb_level);
      sprintf(nlen,"%d",cache_blocking[cb_level-1].cumN);
      sprintf(cname,"cp%d",cb_level);
      generate_level_copyaccumback(mlen,nlen,cname,"Cstride");
    }

    unnest(); 
    nspace(); fprintf(outf,"}\n");    
    
    /* finish up N */
    nspace(); fprintf(outf,"if (n%d < N) {\n",cb_level);
    nest();

    if (top_level && sepaccum) {
      char buff[BUFSIZ];
      sprintf(buff,"((M-m%d)*(N-n%d))",
	      cb_level,cb_level);
      generate_level_zeroaccum(buff);
    }

    generate_level_k_loop(True,cb_level);
    
    nspace();
    if (top_level && sepaccum) {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,%d,N-n%d,ap%d,bp%d,C_accum,Astride,Bstride,(N-n%d));\n",
	      routine_name,cb_level-1,
	      cb_level,
	      cache_blocking[cb_level-1].cumK,
	      cb_level,
	      cb_level,cb_level,cb_level);
    } else {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,%d,N-n%d,ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,
	      cb_level,
	      cache_blocking[cb_level-1].cumK,
	      cb_level,
	      cb_level,cb_level,cb_level,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
    }

    unnest();
    nspace(); fprintf(outf,"}\n");
    
    
    nspace();
    fprintf(outf,"if (k%d < K) {\n",cb_level);
    nest();

    nspace();
    if (top_level && sepaccum) {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,K-k%d,N-n%d,ap%d,bp%d,C_accum,Astride,Bstride,(N-n%d));\n",
	      routine_name,cb_level-1,
	      cb_level,
	      cb_level,
	      cb_level,
	      cb_level,cb_level,cb_level);
    } else {
      fprintf(outf,
	      "%s_l%d_arb_all(M-m%d,K-k%d,N-n%d,ap%d,bp%d,cp%d,Astride,Bstride,Cstride%s);\n",
	      routine_name,cb_level-1,
	      cb_level,
	      cb_level,
	      cb_level,
	      cb_level,cb_level,cb_level,
	      (((alpha == Arb) && !sepaccum) ? ",alpha" : "" ));
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n");

    if (top_level && sepaccum) {
      char mlen[BUFSIZ],nlen[BUFSIZ],cname[BUFSIZ];
      sprintf(mlen,"(M-m%d)",cb_level);
      sprintf(nlen,"(N-n%d)",cb_level);
      sprintf(cname,"cp%d",cb_level);
      generate_level_copyaccumback(mlen,nlen,cname,"Cstride");
    }
    
    unnest();
    nspace(); fprintf(outf,"}\n");
    
    unnest(); 
    nspace(); fprintf(outf,"}\n");    
  }

}


void
generate_level_routines()
{
  int cb_level; /* cache block level */
  int loc_fp_dtype_num;
  if (sepaccum) {
    assert ( cache_blocking_entries > 1 );
    loc_fp_dtype_num = fp_atype_num;
  } else 
    loc_fp_dtype_num = fp_dtype_num;

  for (cb_level=1; cb_level < cache_blocking_entries; cb_level ++) {

    /* generate arbitrary k routine */
    fprintf(outf,"/* Fixed M,N = %d,%d, Arbitrary K L%d-blocked matrix matrix multiply. */\n",
	    cache_blocking[cb_level].cumM,
	    cache_blocking[cb_level].cumN,
	    cb_level-1);
    
    fprintf(outf,"static void\n");
    fprintf(outf,
	    "%s_l%d_arb_k(int K, const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",
	    routine_name,cb_level,fp_type,fp_type,
	    fp_types[loc_fp_dtype_num]);
    if (alpha == Arb && !sepaccum)
      fprintf(outf,", const %s alpha",fp_type);
    fprintf(outf,")\n{\n");
    nest();
    
    if (cb_level == 1)
      generate_core_code(False,Arb_k,loc_fp_dtype_num);
    else
      generate_level_code(False,cb_level,Arb_k,loc_fp_dtype_num);
    unnest();
    fprintf(outf,"}\n\n");  
    
    

    /* generate arbitrary all routine */
    fprintf(outf,"/* Arbitrary M,K,N L%d-blocked matrix matrix multiply. */\n",
	    cb_level-1);
    
    fprintf(outf,"static void\n");
    fprintf(outf,
	    "%s_l%d_arb_all(const int M, const int K, const int N, const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",
	    routine_name,cb_level,fp_type,fp_type,
	    fp_types[loc_fp_dtype_num]);
    if (alpha == Arb && !sepaccum)
      fprintf(outf,", const %s alpha",fp_type);
    fprintf(outf,")\n{\n");
    nest();
    
    if (cb_level == 1)
      generate_core_code(False,Arb_all,loc_fp_dtype_num);
    else
      generate_level_code(False,cb_level,Arb_all,loc_fp_dtype_num);
    unnest();
    fprintf(outf,"}\n\n");  



    /* generate the fixed routine */
    fprintf(outf,"/* Fixed M,K,N = %d,%d,%d L%d-blocked matrix matrix multiply. */\n",
	    cache_blocking[cb_level].cumM,
	    cache_blocking[cb_level].cumK,
	    cache_blocking[cb_level].cumN,
	    cb_level-1);

    fprintf(outf,"static void\n");
    fprintf(outf,
	    "%s_l%d(const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",
	    routine_name,cb_level,fp_type,fp_type,
	    fp_types[loc_fp_dtype_num]);
    if (alpha == Arb && !sepaccum)
      fprintf(outf,", const %s alpha",fp_type);
    fprintf(outf,")\n{\n");
    nest();

    if (cb_level == 1)
      generate_core_code(False,Fixed,loc_fp_dtype_num);
    else 
      generate_level_code(False,cb_level,Fixed,loc_fp_dtype_num);
    
    unnest();
    fprintf(outf,"}\n\n");  

  }

}

void
printMainRoutineHeader()
{
  fprintf(outf,"%s(const int M, const int K, const int N, const %s *const A, const %s *const B, %s *const C, const int Astride, const int Bstride, const int Cstride",routine_name,fp_type,fp_type,fp_dtype);

  if (alpha == Arb)
    fprintf(outf,", const %s alpha",fp_type);
  if (beta == Arb)
    fprintf(outf,", const %s beta",fp_type);
  fprintf(outf,")\n");
}


void
generate_main_routine()
{
  fprintf(outf,"void\n");
  printMainRoutineHeader();
  fprintf(outf,"{\n");

  nest();

  if (cache_blocking_entries == 1)
    generate_core_code(True,Arb_all,fp_dtype_num);
  else {
    generate_level_code(True,cache_blocking_entries,Arb_all,fp_dtype_num);
  }
  unnest();

  fprintf(outf,"}\n");  

}

void
print_date_string()
{
  time_t tloc;
  struct tm*tms;
  char buf[BUFSIZ];
  time(&tloc);
  tms = localtime(&tloc);
  strftime(buf,BUFSIZ,"%A %B %d %Y, %T %Z",tms);
  fprintf(outf,"%s",buf);
}


void
printUsageMsg()
{

  nspace(); 
  fprintf(outf,"**\n");

  nspace(); 
  fprintf(outf,"** Usage:\n");


  nspace(); 
  fprintf(outf,"**    ");
  printMainRoutineHeader();

  nspace(); 
  fprintf(outf,"** where\n");

  nspace(); 
  fprintf(outf,"**  ");
  if (opA == Transpose)
    fprintf(outf,"transpose(A)");
  else 
    fprintf(outf,"A");
  fprintf(outf," is an MxK matrix\n");

  nspace(); 
  fprintf(outf,"**  ");
  if (opB == Transpose)
    fprintf(outf,"transpose(B)");
  else 
    fprintf(outf,"B");
  fprintf(outf," is an KxN matrix\n");

  nspace(); 
  fprintf(outf,"**  ");
  fprintf(outf,"C");
  fprintf(outf," is an MxN matrix\n");

  nspace(); 
  fprintf(outf,"**  Astride is the number of entries between the start of each row of A\n");
  

  nspace(); 
  fprintf(outf,"**  Bstride is the number of entries between the start of each row of B\n");
  

  nspace(); 
  fprintf(outf,"**  Cstride is the number of entries between the start of each row of C\n");

  nspace(); 
  fprintf(outf,"**\n");


}



void
printStaticAccumulator()
{

  fprintf(outf,"\n\n");

  nspace();
  fprintf(outf,"/* Separate Memory used for accumulating the destination blocks.*/\n");
  nspace();
  fprintf(outf,"static %s C_accum[%d*%d];\n",
	  fp_atype,
	  cache_blocking[cache_blocking_entries-1].cumM,
	  cache_blocking[cache_blocking_entries-1].cumN);

  fprintf(outf,"\n\n");
}


void
generate_header(int argc,char *argv[])
{
  int i;

  nspace();
  fprintf(outf,"/*\n");

  nspace();
  fprintf(outf,"**\n");


  nspace();
  fprintf(outf,"** PHiPAC Matrix-Matrix Code for the operation:\n");

  nspace();
  fprintf(outf,"**    C = ");
  if (alpha == Arb)
    fprintf(outf,"alpha*");
  else if (alpha == Fix)
    fprintf(outf,"%s*",alpha_val);
  if (opA == Transpose)
    fprintf(outf,"transpose(A)*");
  else 
    fprintf(outf,"A*");
  if (opB == Transpose)
    fprintf(outf,"transpose(B)");
  else 
    fprintf(outf,"B");
  if (beta == One)
    fprintf(outf," + C");
  else if (beta == Arb)
    fprintf(outf," + beta*C");
  else if (beta == Fix)
    fprintf(outf," + %s*C",beta_val);
  fprintf(outf,"\n");

  nspace();
  fprintf(outf,"**\n");

  nspace();
  fprintf(outf,"** Automatically Generated by mm_gen (%s) using the command:\n",rcsrevision);

  nspace();
  fprintf(outf,"**    %s ",argv[0]);
  for (i=1;i<argc;i++)
    fprintf(outf,"%s ",argv[i]);
  fprintf(outf,"\n");

  nspace();
  fprintf(outf,"**\n");

  nspace();
  fprintf(outf,"** Run '%s -help' for help.\n",argv[0]);

  nspace();
  fprintf(outf,"**\n");

  if (do_fringes == False) {
    nspace();
    fprintf(outf,"** WARNING: Becaues of the -nofringes option, this routine\n");
    nspace();
    fprintf(outf,"** does not handle the register block fringes. That means\n");
    nspace();
    fprintf(outf,"** this routine will only work for matrices with sizes\n");
    nspace();
    fprintf(outf,"** M, K, and N a multiple of %d, %d, and %d respectively.\n",
	    cache_blocking[0].M,cache_blocking[0].K,cache_blocking[0].N);
    nspace();
    fprintf(outf,"**\n");
  }



  nspace();
  fprintf(outf,"** Generated on: "); print_date_string();
  fprintf(outf,"\n");

  nspace();
  fprintf(outf,"** Created by: Jeff Bilmes <bilmes@cs.berkeley.edu>\n");
  nspace();
  fprintf(outf,"**             http://www.icsi.berkeley.edu/~bilmes/phipac\n");

  nspace(); 
  fprintf(outf,"**\n");

  printUsageMsg();

  nspace();
  fprintf(outf,"**\n");

  fprintf(outf,"** \"Copyright (c) 1995 The Regents of the University of California.  All\n");
  fprintf(outf,"** rights reserved.\"  Permission to use, copy, modify, and distribute\n");
  fprintf(outf,"** this software and its documentation for any purpose, without fee, and\n");
  fprintf(outf,"** without written agreement is hereby granted, provided that the above\n");
  fprintf(outf,"** copyright notice and the following two paragraphs appear in all copies\n");
  fprintf(outf,"** of this software.\n");
  fprintf(outf,"**\n");
  fprintf(outf,"** IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR\n");
  fprintf(outf,"** DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT\n");
  fprintf(outf,"** OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF\n");
  fprintf(outf,"** CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n");
  fprintf(outf,"**\n");
  fprintf(outf,"** THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,\n");
  fprintf(outf,"** INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY\n");
  fprintf(outf,"** AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS\n");
  fprintf(outf,"** ON AN \"AS IS\" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO\n");
  fprintf(outf,"** PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.\n");

  nspace();
  fprintf(outf,"**\n");

  nspace();
  fprintf(outf,"*/\n\n");


  if (sepaccum)
    printStaticAccumulator();

}


int main(int argc,char*argv[])
{
  parse_args(argc,argv);

  initialize();
  generate_header(argc,argv);
  generate_macros();

  generate_level_routines();  
  generate_main_routine();
  return 0;
}
