const char *rprf_rcsid = "$Header: /u/spert/src/rprf/RCS/rprf.c,v 1.13 1995/09/19 21:36:52 krste Exp $";
/*
   Routines to help with performance testing.
*/

#include <stdarg.h>
#include <stdlib.h>
#include <float.h>
#include <stdio.h>
#include <string.h>

/* For systems with broken header files. */
/* TODO: Should be handled by configure. */
int fprintf(FILE*, const char*, ...);
int printf(const char*, ...);
long strtol(const char*, char**, int);
double strtod(const char*, char**);

#include "rprf.h"

static void
error(const char *msg)
{
    fprintf(stderr, "rprf: %s\n", msg);
    exit(1);
}

static const char* progname;	/* Name of the current perf program. */
static const char *user_format; /* Format string user provided. */

static void
usage()
{
    fprintf(stderr, "Usage: %s [<options>] [<perf_args>]\n"
	    "where <options> is zero of more of:\n"
	    "  -i <inner_count=100>\n"
	    "  -o <outer_count=1>\n"
	    "  -A <log2_size_of_arena=24>\n"
	    "  -c <log2_size_of_cache=13>\n"
	    "  -a <next_alignment_offset=0>\n"
            "and <perf_args> follow the format string\n"
            "  %s\n"
            "where\n"
            "  i represents an integer\n"
            "  f represents a single-precision floating-point number\n"
            "  d represents a double-precision floating-point number\n"
            "  s represents a string\n",
	    progname,
            user_format);
}

static void
usage_error(const char *msg)
{
    fprintf(stderr, "Error: %s\n", msg);
    usage(progname);
    exit(1);
}

long rprf_n_inner = 100;      /* Default number of repeats. */
long rprf_n_outer = 1;        /* Number of outer loops. */

/* Variables holding argument line info about memory arena. */
/* 16 MB standard arena size. */
#ifndef ARENA_LG2SIZE_DEFAULT
# define ARENA_LG2SIZE_DEFAULT (24)
#endif

static long arena_lg2size = ARENA_LG2SIZE_DEFAULT;

/* 8 KB standard cache size, works as page alignment on most systems. */
#ifndef CACHE_LG2SIZE_DEFAULT
#  define CACHE_LG2SIZE_DEFAULT (13)
#endif

static long cache_lg2size = CACHE_LG2SIZE_DEFAULT;

static char *arenap = NULL;       /* Pointer to memory arena. */
static size_t arena_next = 0;     /* Start of next place to allocate. */
static size_t arena_size = 0;     /* Size of arena. */

/* Quick hack, use simple table to hold list of offsets. */
#ifndef MAX_OFFSETS
#  define MAX_OFFSETS (64)
#endif

static int n_offsets = 0;
static int next_offset = 0;
static long offset[MAX_OFFSETS];

static void
arena_init(void)
{
    /* Work out arena start alignment. */
    const size_t cache_size = 1 << cache_lg2size;
    const size_t cache_mask = cache_size - 1;

    /* Initialize memory arena. */
    arena_next = 0;
    arena_size = (size_t) (1 << arena_lg2size);

    arenap = (char *) malloc(arena_size + cache_mask);
    if (arenap==NULL)
        error("Couldn't allocate memory arena.");

#ifdef DEBUG_PRINTF
    fprintf(stderr, "Original arenap=%08lx, ", (unsigned long) arenap);
#endif

    /* Bump arenap to start on cache-size-aligned boundary. */
    /* HACK: Casting pointer to unsigned long!!! */
    arenap = (char *) (((unsigned long) arenap + cache_mask) & ~ cache_mask);

#ifdef DEBUG_PRINTF
    fprintf(stderr, "aligned arenap=%08lx\n", (unsigned long) arenap);
#endif
}

void *
rprf_malloc(size_t len, size_t alignment)
{
    const size_t align_mask = alignment - 1;
    void * retp;                  /* Value to return. */

    /* Bump index according to offset info. */
    if (next_offset < n_offsets)
    {
        /* First move end to next cache boundary. */
        const size_t cache_size = 1 << cache_lg2size;
        const size_t cache_mask = cache_size - 1;
        arena_next = (arena_next + cache_mask) & ~cache_mask;

        /* Then add on offset. */
        arena_next += offset[next_offset++];
    }
    /* Else abut next block. */

    /* Bump index to next aligned boundary. */
    if (alignment & align_mask)
        error("rprf_malloc(): Alignment not a power of 2.");
    arena_next = (arena_next + align_mask) & ~align_mask;

    /* Store pointer to return for this malloc. */
    retp = (void *) (arenap + arena_next);

    /* Bump index past end. */
    arena_next += len;

    if (arena_next > arena_size)
        error("rprf_malloc(): "
              "No more space in memory arena, increase -A argument.");

#ifdef DEBUG_PRINTF
    fprintf(stderr, "retp=%08lx\n", (unsigned long) retp);
#endif

    return retp;
}

static void
parse_long(const char*const s, long *const lp)
{
    size_t len = strlen(s);
    char *ptr;

    *lp = strtol(s, &ptr, 0);

    if (ptr != (s+len))
        error("Not an integer argument.");
}

static void
parse_float(const char*const s, float *const fp)
{
    size_t len = strlen(s);
    char *ptr;
    double val;
    double absval;

    val = strtod(s, &ptr);

    if (ptr != (s+len))
        error("Not a floating point argument.");

    /* Do fabs() here to avoid linking with math library. */
    absval = (val < 0) ? -val : val;

    if ( (absval != 0.0 && absval < FLT_MIN) || absval > FLT_MAX)
        error("Float argument out of range.");

    *fp = (float) val;
}

static void
parse_double(const char*const s, double *const dp)
{
    size_t len = strlen(s);
    double val;
    char *ptr;

    val = strtod(s, &ptr);

    if (ptr != (s+len))
        error("Not a floating point argument.");

    if (val < -DBL_MAX || val > DBL_MAX)
        error("Float argument out of range.");
    *dp = val;
}

void
rprf_parse_args(int argc, const char* argv[], const char* format, ... )
{
    va_list ap;

    /* Never look at argv[0]. */
    argc--;
    progname = *argv++;
    user_format = format;

    /* First parse standard rprf arguments. */

    while (argc)
    {
        const char *argp = *argv;

        if (argp[0] == '-')
        {
            /* Move to next argument. */
            argc--;
            argv++;

            switch (argp[1])
            {
            case 'i':
                /* Number of inner timing loops. */
                if (argc--)
                    parse_long(*argv++, &rprf_n_inner);
                else
                    usage_error("Argument to -i missing.");
                break;

            case 'o':
                /* Number of outer timing loops. */
                if (argc--)
                    parse_long(*argv++, &rprf_n_outer);
                else
                    usage_error("Argument to -o missing.");
                break;

            case 'A':
                /* Size of memory arena, as power of 2 number of bytes. */
                if (argc--)
                {
                    parse_long(*argv++, &arena_lg2size);

                    if (arena_lg2size < 1 || arena_lg2size > 31)
                        usage_error(
                            "Memory arena size negative or too large.");
                }
                else
                    usage_error("Argument to -A missing.");
                break;

            case 'c':
                /* Size of cache, as power of 2 number of bytes. */
                if (argc--)
                    parse_long(*argv++, &cache_lg2size);
                else
                    usage_error("Argument to -c missing.");
                break;

            case 'a':
                /* Size of next alignment offset, given in bytes. */
                if (argc--)
                {
                    if (n_offsets==MAX_OFFSETS)
                        error("Too many alignment offsets.");

                    parse_long(*argv++, &offset[n_offsets++]);
                }
                else
                    usage_error("Argument to -a missing.");
                break;

            default:
                usage_error("Illegal argument.");
		break;
            }
        }
        else
            break;              /* Switch to format string arguments. */
    }

    /* Now parse any remaining arguments according to format string. */
    va_start(ap, format);

    if (argc < strlen(format))
        usage_error("Not enough <perf_args>.");

    while (argc--)
    {
        long *lp;
        float *fp;
        double *dp;
	const char** sp;

        printf("%s ", *argv);

        switch (*format++)
        {
        case 'i':
            lp = va_arg(ap, long *);
            parse_long(*argv++, lp);
            break;
        case 'f':
            fp = va_arg(ap, float *);
            parse_float(*argv++, fp);
            break;
        case 'd':
            dp = va_arg(ap, double *);
            parse_double(*argv++, dp);
            break;
	case 's':
	    sp = va_arg(ap, const char**);
	    *sp = *argv++;
	    break;
        case '\0':
	    usage_error("Too many <perf_args>.");
            break;
        default:
            error("Format string contains invalid characters.");
            break;
        }
    }

    va_end(ap);

    arena_init();
}
